import random
import time
from typing import Any, Dict, List

from spacy.tokens import Doc

from conspiracies.docprocessing.relationextraction.gptprompting import PromptTemplate
from conspiracies.registry import registry


@registry.prompt_apis.register("conspiracies/openai_gpt3_api")
def create_openai_gpt3_prompt_api(
    prompt_template: PromptTemplate,
    api_key: str,
    model_name: str,
    api_kwargs: Dict[Any, Any],
):
    def openai_prompt(targets: List[str]) -> List[str]:
        """"""
        try:
            import openai
        except ImportError:
            raise ImportError(
                "The OpenAI API requires the openai package to be installed. "
                "You can install the requirements for this module using "
                "`pip install conspiracies[openai]`.",
            )

        openai.api_key = api_key

        responses: List[str] = []
        for target in targets:
            # Run loop until reaching return and response is returned
            while True:
                try:
                    response = openai.Completion.create(
                        model=model_name,
                        prompt=prompt_template.create_prompt(target),
                        **api_kwargs,
                    )
                    responses.append(response["choices"][0]["text"])
                    break

                except openai.error.InvalidRequestError as e:
                    print("Invalid request got error: ", e)
                    print("Retrying with fewer examples...")
                    # Randomly select an example to drop
                    current_examples: List[Doc] = list(prompt_template.examples)
                    current_examples.pop(random.randrange(len(current_examples)))
                    prompt_template.set_examples(current_examples)  # type: ignore
                except openai.error.APIConnectionError:
                    print("Connection reset, waiting 20 sec then retrying...")
                    time.sleep(20)

        return responses

    return openai_prompt


@registry.prompt_apis.register("conspiracies/openai_chatgpt_api")
def create_openai_chatgpt_prompt_api(
    prompt_template: PromptTemplate,
    api_key: str,
    model_name: str,
    api_kwargs: Dict[Any, Any],
):
    def openai_prompt(targets: List[str]) -> List[str]:
        """"""
        try:
            import openai
        except ImportError:
            raise ImportError(
                "The OpenAI API requires the openai package to be installed. "
                "You can install the requirements for this module using "
                "`pip install conspiracies[openai]`.",
            )

        openai.api_key = api_key
        message_example = prompt_template.create_prompt("test")
        assert isinstance(message_example, list) and isinstance(
            message_example[0],
            dict,
        ), "ChatGPT requires a list of message dicts. Consider using chatGPTPromptTemplate as template."  # noqa: E501

        responses: List[str] = []
        for target in targets:
            # Run loop until reaching return and response is returned
            while True:
                try:
                    response = openai.ChatCompletion.create(
                        model=model_name,
                        messages=prompt_template.create_prompt(target),
                        **api_kwargs,
                    )
                    responses.append(response["choices"][0]["message"]["content"])
                    break

                except openai.error.InvalidRequestError:
                    # Randomly select an example to drop
                    current_examples = prompt_template.examples
                    current_examples.pop(random.randrange(len(current_examples)))
                    examples = current_examples
                    prompt_template.set_examples(examples)  # type: ignore

                except openai.error.ConnectionResetError:
                    print("Connection reset, waiting 20 sec then retrying...")
                    time.sleep(20)

        return responses

    return openai_prompt



import networkx as nx
import json
from collections import Counter
import matplotlib.pyplot as plt
import os
import numpy as np
import ndjson
import typing
from typing import Dict, Tuple, Optional, List
from networkx.drawing.layout import (
    fruchterman_reingold_layout,
    kamada_kawai_layout,
)


def load_json(file_path: str) -> dict:
    """Loads a json file.

    Args:
        file_path (str): path to the json file

    Returns:
        data (dict): the data from the file as a dictionary
    """
    with open(file_path, mode="r", encoding="utf8") as f:
        data = json.load(f)
    return data


def most_frequent_tuples(
    node_dict: Dict[int, tuple],
    n: int,
    hard_filter: bool = False,
) -> Dict[int, tuple]:
    """Find the tuples that contain one of the n most frequent elements.

    Args:
        node_dict (dict): dictionary with node ids as keys and node pairs as values
        n (int): The number of most frequent elements to keep

    Returns:
        result (dict): filtered dictionary with node ids as keys and node pairs as
            values
    """
    element_counter: typing.Counter = Counter()
    # Count the frequency for all elements and identify the n most frequent
    for i, tup in node_dict.items():
        for string in tup:
            element_counter[string] += 1
    most_common_strings = set(string for string, _ in element_counter.most_common(n))
    highest_freq = element_counter.most_common(1)[0][1]

    # Only keep the nodes that contain one of the most frequent elements
    result = {}
    for i, tup in node_dict.items():
        if any(string in most_common_strings for string in tup):
            fr = sum([element_counter[tup[0]], element_counter[tup[1]]])
            # If hard_filter, keep only pairs of nodes at least at frequent
            # as the most frequent node
            if hard_filter:
                if fr >= highest_freq:
                    result[i] = tup
            else:
                result[i] = tup
    return result


def get_nodes_edges(
    event: str,
    file: str,
    remove_self_edges: bool = True,
    remove_custom_nodes: Optional[List[str]] = None,
    n_most_frequent: int = 10,
    hard_filter: bool = False,
    save: Optional[str] = None,
) -> Tuple[Dict[int, tuple], Dict[int, tuple]]:
    """Loads nodes and edges from a json file, removes self edges and only
    keeps the n most frequent nodes.

    Args:
        event (str): the name of the event
        file (str): the specific file to load. Must be placed in a folder with the event
            name
        remove_self_edges (bool, optional): Whether or not to remove self edges.
            Defaults to True.
        remove_custom_nodes (list, optional): A list of nodes to remove from the graph.
            Defaults to None.
        n_most_frequent (int, optional): The number of most frequent elements to
            include. Defaults to 10.
        hard_filter (bool, optional): Passed to most_frequent_tuples. Whether or not to
            filter away pairs of nodes that are less frequent than the single most
            frequent. Defaults to False.
        save (str, optional): If provided, the filtered nodes and edges will be saved
            to this path.

    Returns:
        Tuple[Dict[int, tuple], Dict[int, tuple]]: a dictionary with the most frequent
        nodes and a dictionary with the associated edges
    """
    nodes_edges = load_json(
        os.path.join(event, file),
    )
    nodes = {i: tuple(edge) for i, edge in enumerate(nodes_edges["nodes"])}
    edges = {i: edge for i, edge in enumerate(nodes_edges["edges"])}

    # Removing self edges and only keeping the most frequent nodes
    if remove_self_edges:
        nodes = {i: (e1, e2) for i, (e1, e2) in nodes.items() if e1 != e2}
    if remove_custom_nodes:
        nodes = {
            i: (e1, e2)
            for i, (e1, e2) in nodes.items()
            if e1 not in remove_custom_nodes and e2 not in remove_custom_nodes
        }
    most_frequent_nodes = most_frequent_tuples(nodes, n_most_frequent, hard_filter)
    associated_edges = {
        i: edge for i, edge in edges.items() if i in most_frequent_nodes.keys()
    }
    if save:
        with open(save, "w") as f:
            ndjson.dump(
                {
                    "nodes": list(most_frequent_nodes.values()),
                    "edges": list(associated_edges.values()),
                },
                f,
            )
    return most_frequent_nodes, associated_edges


def quantile_min_value(lst, quantile):
    q = np.quantile(lst, quantile)
    return min(filter(lambda x: x >= q, lst))


def min_max_normalize(list_to_normalize: list, min_constant=0.5) -> list:
    """Normalizes a list between 0 and 1 using min-max normalization.

    Args:
        list_to_normalize (list): The list to normalize

    Returns:
        list: The normalized list
    """
    min_value = min(list_to_normalize)
    max_value = max(list_to_normalize)
    if min_value == max_value:
        return list_to_normalize
    scaled_list = [(x - min_value) / (max_value - min_value) for x in list_to_normalize]
    return [x + min_constant for x in scaled_list]


def create_network_graph(
    node_list,
    edge_list,
    title: Optional[str] = None,
    layout=fruchterman_reingold_layout,
    k: float = 0.3,
    node_size_mult: float = 3000,
    node_size_min: float = 0.001,
    edge_weight_mult: float = 5,
    fontsize: int = 12,
    edge_quantile_value: float = 0.90,
    node_quantile_value: Optional[float] = None,
    node_color: str = "#146D25",
    edge_color: str = "#54A463",
    fig_size: int = 10,
    plot_coordinates: bool = False,
    seed: Optional[int] = None,
    draw_labels: bool = True,
    save=False,
):
    G = nx.Graph()
    G.add_edges_from(list(node_list.values()))
    c = Counter(list(node_list.values()))  # edge weights = frequency of edge
    for u, v, d in G.edges(data=True):
        # Make the graph undirected - for some reason, the tuples are sometimes reversed
        # in the edge list
        d["weight"] = c[(u, v)] + c[(v, u)]

    edge_label_weight_cutoff = quantile_min_value(list(c.values()), edge_quantile_value)
    edges_to_draw = {}
    for n, nodes in node_list.items():
        if c[nodes] >= edge_label_weight_cutoff:
            edges_to_draw[nodes] = edge_list[n]

    if layout == kamada_kawai_layout:
        pos = layout(G, scale=2)
    else:
        pos = layout(G, k=k, seed=seed)

    degrees = nx.degree(G)
    normalized_degrees = min_max_normalize(
        [d[1] for d in degrees],
        min_constant=node_size_min,
    )

    plt.figure(figsize=(fig_size, fig_size))
    if title:
        plt.title(
            title,
            color="k",
            fontsize=fontsize + 8,
        )

    edge_weights = min_max_normalize([d["weight"] for _, _, d in G.edges(data=True)])

    nx.draw(
        G,
        pos,
        # node_size=non_norm_degrees,
        node_size=[d * node_size_mult for d in normalized_degrees],
        node_color=node_color,
        edge_color=edge_color + "80",
        # width=[d["weight"] ** edge_weight_mult for _, _, d in G.edges(data=True)],
        width=[e * edge_weight_mult for e in edge_weights],
    )
    if draw_labels:
        nx.draw_networkx_edge_labels(
            G,
            pos,
            edge_labels=edges_to_draw,
            font_size=fontsize - 1,
            label_pos=0.5,
            bbox=dict(
                facecolor="white",
                edgecolor=edge_color,
                alpha=0.8,
                boxstyle="round,pad=0.2",
            ),
        )

        offset = 0.015
        for node, (x, y) in pos.items():
            h_align = "center"
            v_align = "center"
            if x < 0:
                x -= offset
                h_align = "right"
            if x > 0:
                x += offset
                h_align = "left"
            if y < 0:
                y -= offset
                v_align = "top"
            if y > 0:
                y += offset
                v_align = "bottom"
            if plot_coordinates:
                label = f"{node} ({x:.2f}, {y:.2f})"
            else:
                label = node
            if node_quantile_value:
                node_label_draw_cutoff = quantile_min_value(
                    [value for key, value in degrees],
                    node_quantile_value,
                )
                if degrees[node] >= node_label_draw_cutoff:
                    plt.text(
                        x,
                        y,
                        label,
                        fontsize=fontsize,
                        color="k",
                        ha=h_align,
                        va=v_align,
                        bbox=dict(
                            facecolor="white",
                            edgecolor=node_color,
                            alpha=0.8,
                            boxstyle="round,pad=0.1",
                        ),
                    )
            else:
                plt.text(
                    x,
                    y,
                    label,
                    fontsize=fontsize,
                    # fontname="Helvetica",
                    color="k",
                    ha=h_align,
                    va=v_align,
                    bbox=dict(
                        facecolor="white",
                        edgecolor=node_color,
                        alpha=0.8,
                        boxstyle="round,pad=0.1",
                    ),
                )
    if save:
        plt.savefig(
            f"{save}.png",
            format="PNG",
            bbox_inches="tight",
        )
    return G



from spacy.tokens import Span
import time
import os
from pathlib import Path
from typing import List, Generator
import spacy
from transformers import AutoTokenizer
import argparse

# Conspiracies
from conspiracies.preproc import wordpiece_length_normalization
from extract_utils import load_ndjson, write_txt


def build_coref_pipeline():
    nlp_coref = spacy.blank("da")
    nlp_coref.add_pipe("sentencizer")
    nlp_coref.add_pipe("allennlp_coref")

    return nlp_coref


def yield_one_article(articles: List[dict]) -> Generator:
    for article in articles:
        yield article["text"]


def exclude_newspapers(
    files: List[str],
    papers_to_exclude: List[str] = ["weekendavisen"],
) -> List[str]:
    """Excludes newspapers from file list
    Args:
        file_list (List[str]): list of files
        papers_to_exclude (List[str]): list of newspapers to exclude
    Returns:
        file_list (List[str]): list of files without excluded newspapers
    """
    file_list = [
        file for file in files if not any(paper in file for paper in papers_to_exclude)
    ]
    return file_list


nlp_for_normalization = spacy.load("da_core_news_sm")
tokenizer = AutoTokenizer.from_pretrained("vesteinn/DanskBERT")


def process_file(
    file: str,
    event: str,
    norm_nlp=nlp_for_normalization,
    tokenizer=tokenizer,
):
    start = time.time()
    data: List[dict] = load_ndjson(file)

    # Create pipeline to resolve coreference
    spacy.prefer_gpu()
    nlp_coref = spacy.blank("da")
    nlp_coref.add_pipe("sentencizer")
    nlp_coref.add_pipe("allennlp_coref")

    # Create pipeline to extract relations
    nlp = spacy.load("da_core_news_sm")
    nlp.add_pipe("sentencizer")
    nlp.add_pipe(
        "heads_extraction",
        config={"normalize_to_entity": True, "normalize_to_noun_chunk": True},
    )
    config = {"confidence_threshold": 2.7, "model_args": {"batch_size": 10}}
    nlp.add_pipe("relation_extractor", config=config)

    for i, article in enumerate(yield_one_article(data)):
        normalized_article = list(
            wordpiece_length_normalization(
                [article],
                norm_nlp,
                tokenizer,
                max_length=500,
            ),
        )
        if i % 10 == 0:
            print(f"Now processing article {i}/{len(data)}")

        # Resolve coreference
        coref_docs = nlp_coref.pipe(normalized_article)
        resolved_docs = (d._.resolve_coref for d in coref_docs)

        # Extract relations
        docs = nlp.pipe(resolved_docs)

        # Extract triplets and write to file
        subjects, predicates, objects, triplets = [], [], [], []
        for doc in docs:
            for triplet in doc._.relation_triplets:
                if not all(isinstance(element, Span) for element in triplet):
                    continue
                if len(triplet) != 3:
                    continue
                subject = triplet[0]._.most_common_ancestor.text
                predicate = triplet[1]._.most_common_ancestor.text
                object = triplet[2]._.most_common_ancestor.text
                subjects.append(subject)
                predicates.append(predicate)
                objects.append(object)
                triplets.append((subject, predicate, object))
        if len(triplets) == 0:
            continue
        write_txt(
            os.path.join("extracted_triplets_papers", event, "subjects.txt"),
            subjects,
            "a+",
        )
        write_txt(
            os.path.join("extracted_triplets_papers", event, "predicates.txt"),
            predicates,
            "a+",
        )
        write_txt(
            os.path.join("extracted_triplets_papers", event, "objects.txt"),
            objects,
            "a+",
        )
        write_txt(
            os.path.join("extracted_triplets_papers", event, "triplets.txt"),
            triplets,
            "a+",
        )
    write_txt(
        os.path.join("extracted_triplets_papers", event, "processed_files.txt"),
        [file],
        "a+",
    )
    print(f"Finished piping {file} in {time.time() - start} seconds")


def main(event: str, papers_to_exclude: List[str]):
    event_path = os.path.join(
        "/home",
        os.getlogin(),
        "narratives-political-events",
        "data",
        "clean_news",
        event,
    )
    Path(os.path.join("extracted_triplets_papers", event)).mkdir(
        parents=True,
        exist_ok=True,
    )
    start = time.time()
    files = os.listdir(event_path)
    # Do not look at articles from specified papers. Default is weekendavisen
    files = exclude_newspapers(files, papers_to_exclude)
    print("starting processing all articles")
    for file_n, file in enumerate(files):
        print(
            f"\n__________________________________________\nNow processing {file} ({file_n+1}/{len(files)})",
        )
        file_path = os.path.join(event_path, file)
        process_file(file_path, event)

    print(
        f"Finished processing all articles in {time.time() - start} seconds. Now writing to file",
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-e",
        "--event",
        type=str,
        help="Name of the event to run triplet extraction on",
    )
    parser.add_argument(
        "-ex",
        "--exclude",
        type=str,
        nargs="+",
        default=["weekendavisen"],
        help="Newspapers to exclude from the extraction",
    )

    args = parser.parse_args()
    main_start = time.time()
    main(args.event, args.exclude)



import pytest
import torch

from .utils import nlp_da  # noqa F401


@pytest.mark.skip(reason="Avoid downloading the model on GitHub actions")
def test_relationextraction_component_pipe(nlp_da):  # noqa F811
    test_sents = [
        "Pernille Blume vinder delt EM-sølv i Ungarn.",
        "Pernille Blume blev nummer to ved EM på langbane i disciplinen 50 meter fri.",
    ]

    # change these to your purposes. 2.5 is the default confidence threshold(the bulk of
    # bad relations not kept and the majority of correct ones kept)
    # batch_size should be changed according to your device. Can most likely be bumped
    # up a fair bit
    config = {"confidence_threshold": 2.5, "model_args": {"batch_size": 10}}
    nlp_da.add_pipe("relation_extractor", config=config)

    pipe = nlp_da.pipe(test_sents)

    for d in pipe:
        print(d.text, "\n", d._.relation_triplets)


@pytest.mark.skip(reason="Avoid downloading the model on GitHub actions")
def test_relationextraction_component_pipe_multiprocessing(nlp_da):  # noqa F811
    test_sents = [
        "Pernille Blume vinder delt EM-sølv i Ungarn.",
        "Pernille Blume blev nummer to ved EM på langbane i disciplinen 50 meter fri.",
    ] * 5

    nlp_da.add_pipe("relation_extractor")

    # multiprocessing and torch with multiple threads result in a deadlock, therefore:
    torch.set_num_threads(1)

    pipe = nlp_da.pipe(test_sents, n_process=2, batch_size=5)

    for d in pipe:
        print(d.text, "\n", d._.relation_triplets)


@pytest.mark.skip(reason="Avoid downloading the model on GitHub actions")
def test_relation_extraction_component_single(nlp_da):  # noqa F811
    nlp_da.add_pipe("relation_extractor", config={"confidence_threshold": 1.8})
    doc = nlp_da("Obama is the former president of the United States.")
    triplet_str = [
        tuple([str(t[1]) for t in triplet]) for triplet in doc._.relation_triplets
    ]
    assert triplet_str == [
        ("Obama", "is", "the former president of the United States"),
    ]


@pytest.mark.skip(reason="Avoid downloading the model on GitHub actions")
def test_relation_extraction_multi_sentence(nlp_da):  # noqa F811
    nlp_da.add_pipe("relation_extractor")
    doc = nlp_da(
        "Barack Obama is the former president of the United States. Du har to heste"
        + " der hedder Jens og Frode",
    )
    triplet_str = [
        tuple([str(t[1]) for t in triplet]) for triplet in doc._.relation_triplets
    ]
    assert triplet_str == [
        ("Barack Obama", "is", "the former president of the United States"),
        ("Du", "har to heste", "der hedder Jens og Frode"),
    ]


@pytest.mark.skip(reason="Avoid downloading the model on GitHub actions")
def test_relation_extraction_empty_string(nlp_da):  # noqa F811
    nlp_da.add_pipe("relation_extractor")
    doc = nlp_da("")
    assert len(doc._.relation_triplets) == 0


@pytest.mark.skip(reason="Avoid downloading the model on GitHub actions")
def test_relation_extraction_no_extracted_relation(nlp_da):  # noqa F811
    nlp_da.add_pipe("relation_extractor")
    doc = nlp_da("Ingen relation")
    assert len(doc._.relation_triplets) == 0



import pytest
from conspiracies.docprocessing.relationextraction.gptprompting import (
    MarkdownPromptTemplate1,
    MarkdownPromptTemplate2,
    PromptTemplate1,
    PromptTemplate2,
    XMLStylePromptTemplate,
    chatGPTPromptTemplate,
)

from .test_data.prompt_data import (
    MarkdownPromptTemplate1_expected_prompt,
    MarkdownPromptTemplate2_expected_prompt,
    PromptTemplate1_expected_prompt,
    PromptTemplate2_expected_prompt,
    XMLStylePromptTemplate_expected_prompt,
    chatGPTPromptTemplate_expected_prompt,
    load_examples,
    test_tweet,
)


def get_examples_task_introduction():
    task_description = "This is a test task description"
    examples = load_examples()
    return examples, task_description, test_tweet


examples, task_description, test_tweet = get_examples_task_introduction()  # noqa


@pytest.mark.parametrize(
    "Template, target, examples, task_description, expected_prompt",
    [
        (
            PromptTemplate1,
            test_tweet,
            examples,
            task_description,
            PromptTemplate1_expected_prompt,
        ),
        (
            PromptTemplate2,
            test_tweet,
            examples,
            task_description,
            PromptTemplate2_expected_prompt,
        ),
        (
            MarkdownPromptTemplate1,
            test_tweet,
            examples,
            task_description,
            MarkdownPromptTemplate1_expected_prompt,
        ),
        (
            MarkdownPromptTemplate2,
            test_tweet,
            examples,
            task_description,
            MarkdownPromptTemplate2_expected_prompt,
        ),
        (
            XMLStylePromptTemplate,
            test_tweet,
            examples,
            task_description,
            XMLStylePromptTemplate_expected_prompt,
        ),
        (
            chatGPTPromptTemplate,
            test_tweet,
            examples,
            task_description,
            chatGPTPromptTemplate_expected_prompt,
        ),
    ],
)
def test_PromptTemplate_create_prompt(
    Template,
    target,
    examples,
    task_description,
    expected_prompt,
):
    template = Template(task_description=task_description, examples=examples)
    prompt = template.create_prompt(target)
    assert prompt == expected_prompt



import argparse


from conspiracies.pipeline.config import PipelineConfig
from conspiracies.pipeline.pipeline import Pipeline


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "project_name",
        nargs="?",
        default=None,
        help="Name of your project under which various output files will be output"
        "under output root set in config (the default for which is output/).",
    )
    arg_parser.add_argument(
        "input_path",
        nargs="?",
        default=None,
        help="Input path for preprocessing of documents. Can be a glob path, e.g."
        "path/to/files/*.txt. Be mindful of quotes for glob paths.",
    )
    arg_parser.add_argument(
        "-c",
        "--config",
        default="config/default.toml",
        help="Path to configuration file. Refer to config/template.toml for contents.",
    )
    args = arg_parser.parse_args()

    cli_args = {
        "base.project_name": args.project_name,
        "preprocessing.input_path": args.input_path,
    }
    config = PipelineConfig.from_toml_file(args.config, cli_args)

    pipeline = Pipeline(config)
    pipeline.run()



import ndjson
from typing import List, Union, Tuple
from glob import glob


def load_ndjson(path: str) -> List[dict]:
    """Loads ndjson file
    Args:
        path (str): path to ndjson file
    Returns:
        data (dict): data from ndjson file
    """
    with open(path, "r") as f:
        data = ndjson.load(f)
    return data


def ndjson_gen(filepath: str):
    """Creates a generator for ndjson files and yields the tweets."""
    for file in glob(filepath):
        with open(file) as f:
            reader = ndjson.reader(f)
            for post in reader:
                yield post


def write_txt(
    path: str,
    data: Union[List[str], List[Tuple[str, str, str]]],
    method="a",
):
    """Writes data to txt file. Can take either a list of strings or a list of tuples.
    List of strings is e.g. list of subjects, predicates, or objects.
    List of tuples is e.g. list of triplets.
    Args:
        path (str): path to txt file
        data (Union[List[str], List[tuple]]): data to write to file
        method (str, optional): method to use when writing data. Defaults to "a" (i.e., append).

    Returns:
        None
    """
    with open(path, method) as f:
        if isinstance(data[0], tuple):
            f.write("\n".join([", ".join(triplet) for triplet in data]))
        elif isinstance(data[0], str):
            f.write("\n".join(data))  # type: ignore
        f.write("\n")



from typing import List

import spacy

from conspiracies.docprocessing.relationextraction.data_classes import (
    install_extensions,
)
from conspiracies.docprocessing.relationextraction.gptprompting import (
    DocTriplets,
    SpanTriplet,
    StringTriplet,
)
from spacy.tokens import Doc

test_thread = """@user2: I was hurt. END
@user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough. END
"""  # noqa: E501

test_tweet = "@user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough."  # noqa: E501

### For testing prompt parsing ###

PromptTemplate1_expected_response = "\n\n(This) (is) (a test tweet)\n(I) (am commenting)\n(this) (is not) (good enough.)"  # noqa: E501
PromptTemplate1_expected_triplets = [
    StringTriplet(
        subject="This",
        predicate="is",
        object="a test tweet",
        text=test_tweet,
    ),
    StringTriplet(
        subject="this",
        predicate="is not",
        object="good enough.",
        text=test_tweet,
    ),
]

PromptTemplate2_expected_response = "\n\n(This) (is) (a test tweet)\n(I) (am commenting)\n(this) (is not) (good enough.)"  # noqa: E501
PromptTemplate2_expected_triplets = [
    StringTriplet(
        subject="This",
        predicate="is",
        object="a test tweet",
        text=test_tweet,
    ),
    StringTriplet(
        subject="this",
        predicate="is not",
        object="good enough.",
        text=test_tweet,
    ),
]

MarkdownPromptTemplate1_expected_response = """ This | is | a test tweet |
| | I | am commenting | on something someone else said |
| | this | is not | good enough. |
| | This | is not |"""
MarkdownPromptTemplate1_expected_triplets = [
    StringTriplet(
        subject="This",
        predicate="is",
        object="a test tweet",
        text=test_tweet,
    ),
    StringTriplet(
        subject="I",
        predicate="am commenting",
        object="on something someone else said",
        text=test_tweet,
    ),
    StringTriplet(
        subject="this",
        predicate="is not",
        object="good enough.",
        text=test_tweet,
    ),
]

MarkdownPromptTemplate2_expected_response = """| This | is | a test tweet |
| I | am commenting | on something someone else said |
| @user2 | this is not good enough |"""
MarkdownPromptTemplate2_expected_triplets = [
    StringTriplet(
        subject="This",
        predicate="is",
        object="a test tweet",
        text=test_tweet,
    ),
    StringTriplet(
        subject="I",
        predicate="am commenting",
        object="on something someone else said",
        text=test_tweet,
    ),
]

XMLStylePromptTemplate_expected_response = "@user1: @user2 <subject-1>This</subject-1> <predicate-1>is</predicate-1> <object-1>a test tweet</object-1>, <subject-2>I</subject-2> <predicate-2>am commenting</predicate-2> <object-2>on something someone else said</object-2>. @user2 <subject-3>this</subject-3> <predicate-3>is not</predicate-3> <object-3>good enough.</object-3>"  # noqa: E501

XMLStylePromptTemplate_expected_triplets = [
    StringTriplet(
        subject="This",
        predicate="is",
        object="a test tweet",
        subject_char_span=(15, 19),
        predicate_char_span=(20, 22),
        object_char_span=(23, 35),
        text=test_tweet,
    ),
    StringTriplet(
        subject="I",
        predicate="am commenting",
        object="on something someone else said",
        subject_char_span=(37, 38),
        predicate_char_span=(39, 52),
        object_char_span=(53, 83),
        text=test_tweet,
    ),
    StringTriplet(
        subject="this",
        predicate="is not",
        object="good enough.",
        subject_char_span=(92, 96),
        predicate_char_span=(97, 103),
        object_char_span=(104, 116),
        text=test_tweet,
    ),
]

chatGPTPromptTemplate_expected_response = """I extracted the following triplets:
	@user1 - commenting on - something someone else said
	@user2 - this is - not good enough. 
Note that the verb phrase "commenting on" includes the particles "on" and "something else"."""  # noqa: E501

chatGPTPromptTemplate_expected_triplets = [
    StringTriplet(
        subject="@user1",
        predicate="commenting on",
        object="something someone else said",
        text=test_tweet,
    ),
    StringTriplet(
        subject="@user2",
        predicate="this is",
        object="not good enough.",
        text=test_tweet,
    ),
]


def load_gold_triplets() -> List[Doc]:
    """Load two examples of docs with gold annotations in the.

    ._.relation_triplets attribute.
    """
    nlp = spacy.blank("da")
    nlp.add_pipe("sentencizer")
    # gold annotations
    doc = nlp(test_tweet)  # type: ignore

    span_triplets_ = [
        SpanTriplet.from_doc(t, doc=doc, nlp=nlp)  # type: ignore
        for t in XMLStylePromptTemplate_expected_triplets
    ]
    span_triplets = [triplet for triplet in span_triplets_ if triplet is not None]

    if not Doc.has_extension("relation_triplets"):
        install_extensions()
    doc._.relation_triplets = DocTriplets(span_triplets=span_triplets, doc=doc)

    # copy them to test with multiple examples.
    return [doc, doc]


### For testing prompt creation ###

example_tweet_1 = "@Politician1: @minister @journalist @Politician1 er uenig i det I siger. Man kan ikke bare gøre hvad der passer en. Almindelig sund fornuft er det eneste @Politician1 forlanger."  # noqa: E501
example_tweet_2 = "@minister: @minister skriver her en kort kommentar en kort kommentar skyder med skarpt på @Politician1."  # noqa: E501
example_tweet_3 = "@almindelig123: At blande sig i debatten er ikke en god idé. Nogle gange er det bedre at holde sig for sig selv. Men sådan er det, det er der ikke noget at gøre ved."  # noqa: E501
example_tweet_4 = """@fagperson: @Politician1 @minister @journalist Som fagperson vil @fagperson gerne understrege at mine inputs burde fylde mere. 
Det er vigtigt at lytte til hvad @fagperson siger, og I er alle myndighedspersoner der burde vægte sådanne faglige indspark ekstra højt."""  # noqa: E501
example_tweet_5 = "@minister: @minister lytter til @fagperson og @almindelig123. @minister er enig med dem. Solid argumentation har overbevist mig om at flertallet nok har ret."  # noqa: E501

example_triplets_1 = [
    ["@Politician1", "er uenig", "i det I siger"],
    ["@Politician1", "forlanger", "almindelig sund fornuft"],
]
example_triplets_2 = [
    ["@minister", "skriver", "en kort kommentar"],
    ["en kort kommentar", "skyder", "med skarpt"],
]
example_triplets_3 = [
    ["At blande sig", "er ikke", "en god idé"],
    ["det", "er", "bedre at holde sig for sig selv"],
    ["det", "er", "der er ikke noget at gøre ved"],
]
example_triplets_4 = [
    ["@fagperson", "vil", "gerne understrege"],
    ["mine inputs", "burde fylde", "mere"],
    ["det", "er vigtigt", "at lytte til hvad @fagperson siger"],
    ["myndighedspersoner", "burde vægte", "sådanne faglige indspark ekstra højt"],
]
example_triplets_5 = [
    ["@minister", "lytter", "til @fagperson og @almindelig123"],
    ["@minister", "er enig", "med dem"],
    ["Solid argumentation", "har overbevist", "mig om at flertallet nok har ret"],
]


def load_examples() -> List[Doc]:
    """Load five examples of docs with gold triplet annotations.

    The examples are made up mock tweets.
    """
    nlp = spacy.blank("da")
    nlp.add_pipe("sentencizer")

    if not Doc.has_extension("relation_triplets"):
        install_extensions(force=True)
    examples: List[Doc] = []
    for example, triplet_list in [
        (example_tweet_1, example_triplets_1),
        (example_tweet_2, example_triplets_2),
        (example_tweet_3, example_triplets_3),
        (example_tweet_4, example_triplets_4),
        (example_tweet_5, example_triplets_5),
    ]:
        doc = nlp(example)
        examples.append(doc)
        span_triplets_ = [
            SpanTriplet.from_doc(t, doc=doc, nlp=nlp)  # type: ignore
            for t in triplet_list
        ]
        span_triplets = [triplet for triplet in span_triplets_ if triplet is not None]
        doc._.relation_triplets = span_triplets
    return examples


PromptTemplate1_expected_prompt = """This is a test task description

---

Tweet: @Politician1: @minister @journalist @Politician1 er uenig i det I siger. Man kan ikke bare gøre hvad der passer en. Almindelig sund fornuft er det eneste @Politician1 forlanger.
Triplets: (@Politician1) (er uenig) (i det I siger)
\t(@Politician1) (forlanger) (Almindelig sund fornuft)
---
Tweet: @minister: @minister skriver her en kort kommentar en kort kommentar skyder med skarpt på @Politician1.
Triplets: (@minister) (skriver) (en kort kommentar)
\t(en kort kommentar) (skyder) (med skarpt)
---
Tweet: @almindelig123: At blande sig i debatten er ikke en god idé. Nogle gange er det bedre at holde sig for sig selv. Men sådan er det, det er der ikke noget at gøre ved.
Triplets: (At blande sig) (er ikke) (en god idé)
\t(det) (er) (bedre at holde sig for sig selv)
---
Tweet: @fagperson: @Politician1 @minister @journalist Som fagperson vil @fagperson gerne understrege at mine inputs burde fylde mere. 
Det er vigtigt at lytte til hvad @fagperson siger, og I er alle myndighedspersoner der burde vægte sådanne faglige indspark ekstra højt.
Triplets: (@fagperson) (vil) (gerne understrege)
\t(mine inputs) (burde fylde) (mere)
\t(Det) (er vigtigt) (at lytte til hvad @fagperson siger)
\t(myndighedspersoner) (burde vægte) (sådanne faglige indspark ekstra højt)
---
Tweet: @minister: @minister lytter til @fagperson og @almindelig123. @minister er enig med dem. Solid argumentation har overbevist mig om at flertallet nok har ret.
Triplets: (@minister) (lytter) (til @fagperson og @almindelig123)
\t(@minister) (er enig) (med dem)
\t(Solid argumentation) (har overbevist) (mig om at flertallet nok har ret)
---

Tweet: @user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough.
Triplets:"""  # noqa: E501

PromptTemplate2_expected_prompt = """This is a test task description

---

Tweet: @Politician1: @minister @journalist @Politician1 er uenig i det I siger. Man kan ikke bare gøre hvad der passer en. Almindelig sund fornuft er det eneste @Politician1 forlanger.

Tweet: @minister: @minister skriver her en kort kommentar en kort kommentar skyder med skarpt på @Politician1.

Tweet: @almindelig123: At blande sig i debatten er ikke en god idé. Nogle gange er det bedre at holde sig for sig selv. Men sådan er det, det er der ikke noget at gøre ved.

Tweet: @fagperson: @Politician1 @minister @journalist Som fagperson vil @fagperson gerne understrege at mine inputs burde fylde mere. 
Det er vigtigt at lytte til hvad @fagperson siger, og I er alle myndighedspersoner der burde vægte sådanne faglige indspark ekstra højt.

Tweet: @minister: @minister lytter til @fagperson og @almindelig123. @minister er enig med dem. Solid argumentation har overbevist mig om at flertallet nok har ret.


Triplets: (@Politician1) (er uenig) (i det I siger)
\t(@Politician1) (forlanger) (Almindelig sund fornuft)
Triplets: (@minister) (skriver) (en kort kommentar)
\t(en kort kommentar) (skyder) (med skarpt)
Triplets: (At blande sig) (er ikke) (en god idé)
\t(det) (er) (bedre at holde sig for sig selv)
Triplets: (@fagperson) (vil) (gerne understrege)
\t(mine inputs) (burde fylde) (mere)
\t(Det) (er vigtigt) (at lytte til hvad @fagperson siger)
\t(myndighedspersoner) (burde vægte) (sådanne faglige indspark ekstra højt)
Triplets: (@minister) (lytter) (til @fagperson og @almindelig123)
\t(@minister) (er enig) (med dem)
\t(Solid argumentation) (har overbevist) (mig om at flertallet nok har ret)

---

Tweet: @user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough.

Triplets:"""  # noqa: E501

MarkdownPromptTemplate1_expected_prompt = """This is a test task description
| Tweet | Subject | Predicate | Object |
| --- | --- | --- | --- |
| @Politician1: @minister @journalist @Politician1 er uenig i det I siger. Man kan ikke bare gøre hvad der passer en. Almindelig sund fornuft er det eneste @Politician1 forlanger. | @Politician1 | er uenig | i det I siger |
| | @Politician1 | forlanger | Almindelig sund fornuft |
| @minister: @minister skriver her en kort kommentar en kort kommentar skyder med skarpt på @Politician1. | @minister | skriver | en kort kommentar |
| | en kort kommentar | skyder | med skarpt |
| @almindelig123: At blande sig i debatten er ikke en god idé. Nogle gange er det bedre at holde sig for sig selv. Men sådan er det, det er der ikke noget at gøre ved. | At blande sig | er ikke | en god idé |
| | det | er | bedre at holde sig for sig selv |
| @fagperson: @Politician1 @minister @journalist Som fagperson vil @fagperson gerne understrege at mine inputs burde fylde mere. 
Det er vigtigt at lytte til hvad @fagperson siger, og I er alle myndighedspersoner der burde vægte sådanne faglige indspark ekstra højt. | @fagperson | vil | gerne understrege |
| | mine inputs | burde fylde | mere |
| | Det | er vigtigt | at lytte til hvad @fagperson siger |
| | myndighedspersoner | burde vægte | sådanne faglige indspark ekstra højt |
| @minister: @minister lytter til @fagperson og @almindelig123. @minister er enig med dem. Solid argumentation har overbevist mig om at flertallet nok har ret. | @minister | lytter | til @fagperson og @almindelig123 |
| | @minister | er enig | med dem |
| | Solid argumentation | har overbevist | mig om at flertallet nok har ret |
| @user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough. |"""  # noqa: E501

MarkdownPromptTemplate2_expected_prompt = """This is a test task description

@Politician1: @minister @journalist @Politician1 er uenig i det I siger. Man kan ikke bare gøre hvad der passer en. Almindelig sund fornuft er det eneste @Politician1 forlanger.

| Subject | Predicate | Object |
| --- | --- | --- |
| @Politician1 | er uenig | i det I siger |
| @Politician1 | forlanger | Almindelig sund fornuft |

@minister: @minister skriver her en kort kommentar en kort kommentar skyder med skarpt på @Politician1.

| Subject | Predicate | Object |
| --- | --- | --- |
| @minister | skriver | en kort kommentar |
| en kort kommentar | skyder | med skarpt |

@almindelig123: At blande sig i debatten er ikke en god idé. Nogle gange er det bedre at holde sig for sig selv. Men sådan er det, det er der ikke noget at gøre ved.

| Subject | Predicate | Object |
| --- | --- | --- |
| At blande sig | er ikke | en god idé |
| det | er | bedre at holde sig for sig selv |

@fagperson: @Politician1 @minister @journalist Som fagperson vil @fagperson gerne understrege at mine inputs burde fylde mere. 
Det er vigtigt at lytte til hvad @fagperson siger, og I er alle myndighedspersoner der burde vægte sådanne faglige indspark ekstra højt.

| Subject | Predicate | Object |
| --- | --- | --- |
| @fagperson | vil | gerne understrege |
| mine inputs | burde fylde | mere |
| Det | er vigtigt | at lytte til hvad @fagperson siger |
| myndighedspersoner | burde vægte | sådanne faglige indspark ekstra højt |

@minister: @minister lytter til @fagperson og @almindelig123. @minister er enig med dem. Solid argumentation har overbevist mig om at flertallet nok har ret.

| Subject | Predicate | Object |
| --- | --- | --- |
| @minister | lytter | til @fagperson og @almindelig123 |
| @minister | er enig | med dem |
| Solid argumentation | har overbevist | mig om at flertallet nok har ret |

@user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough.

| Subject | Predicate | Object |
| --- | --- | --- |
"""  # noqa: E501

XMLStylePromptTemplate_expected_prompt = """This is a test task description

@Politician1: @minister @journalist @Politician1 er uenig i det I siger. Man kan ikke bare gøre hvad der passer en. Almindelig sund fornuft er det eneste @Politician1 forlanger.
@Politician1: @minister @journalist <subject-1>@Politician1 </subject-1><predicate-1>er uenig </predicate-1><object-1>i det I siger</object-1>. Man kan ikke bare gøre hvad der passer en. <object-2>Almindelig sund fornuft </object-2>er det eneste <subject-2>@Politician1 </subject-2><predicate-2>forlanger</predicate-2>.

@minister: @minister skriver her en kort kommentar en kort kommentar skyder med skarpt på @Politician1.
@minister: <subject-1>@minister </subject-1><predicate-1>skriver </predicate-1>her <object-1>en kort kommentar </object-1><subject-2>en kort kommentar </subject-2><predicate-2>skyder </predicate-2><object-2>med skarpt </object-2>på @Politician1.

@almindelig123: At blande sig i debatten er ikke en god idé. Nogle gange er det bedre at holde sig for sig selv. Men sådan er det, det er der ikke noget at gøre ved.
@almindelig123: <subject-1>At blande sig </subject-1>i debatten <predicate-1>er ikke </predicate-1><object-1>en god idé</object-1>. Nogle gange <predicate-2>er </predicate-2><subject-2>det </subject-2><object-2>bedre at holde sig for sig selv</object-2>. Men sådan er det, det er der ikke noget at gøre ved.

@fagperson: @Politician1 @minister @journalist Som fagperson vil @fagperson gerne understrege at mine inputs burde fylde mere. 
Det er vigtigt at lytte til hvad @fagperson siger, og I er alle myndighedspersoner der burde vægte sådanne faglige indspark ekstra højt.
<subject-1>@fagperson</subject-1>: @Politician1 @minister @journalist Som fagperson <predicate-1>vil </predicate-1>@fagperson <object-1>gerne understrege </object-1>at <subject-2>mine inputs </subject-2><predicate-2>burde fylde </predicate-2><object-2>mere</object-2>. 
<subject-3>Det </subject-3><predicate-3>er vigtigt </predicate-3><object-3>at lytte til hvad @fagperson siger</object-3>, og I er alle <subject-4>myndighedspersoner </subject-4>der <predicate-4>burde vægte </predicate-4><object-4>sådanne faglige indspark ekstra højt</object-4>.

@minister: @minister lytter til @fagperson og @almindelig123. @minister er enig med dem. Solid argumentation har overbevist mig om at flertallet nok har ret.
@minister: <subject-1>@minister </subject-1><predicate-1>lytter </predicate-1><object-1>til @fagperson og @almindelig123</object-1>. <subject-2>@minister </subject-2><predicate-2>er enig </predicate-2><object-2>med dem</object-2>. <subject-3>Solid argumentation </subject-3><predicate-3>har overbevist </predicate-3><object-3>mig om at flertallet nok har ret</object-3>.

@user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough.
"""  # noqa: E501


chatGPTPromptTemplate_expected_prompt = [
    {
        "role": "system",
        "content": "You are a helpful assistant who is also an expert linguist. ",
    },
    {"role": "user", "content": "Hi, I need your help on a linguistic question."},
    {"role": "assistant", "content": "Sure, what is the task?"},
    {
        "role": "user",
        "content": "This is a test task description",
    },
    {
        "role": "assistant",
        "content": "That sounds like something I can help you with. \
Let me try the first example!",
    },
    {
        "role": "user",
        "content": "The tweet is:\n@Politician1: @minister @journalist @Politician1 er uenig i det I siger. Man kan ikke bare gøre hvad der passer en. Almindelig sund fornuft er det eneste @Politician1 forlanger.\n",  # noqa: E501
    },
    {
        "role": "assistant",
        "content": "I extracted the following triplets:\n\t@Politician1 - er uenig - i det I siger\n\t@Politician1 - forlanger - Almindelig sund fornuft\n",  # noqa: E501
    },
    {
        "role": "user",
        "content": "Correct! The next tweet is:\n@minister: @minister skriver her en kort kommentar en kort kommentar skyder med skarpt på @Politician1.\n",  # noqa: E501
    },
    {
        "role": "assistant",
        "content": "I extracted the following triplets:\n\t@minister - skriver - en kort kommentar\n\ten kort kommentar - skyder - med skarpt\n",  # noqa: E501
    },
    {
        "role": "user",
        "content": "Correct! The next tweet is:\n@almindelig123: At blande sig i debatten er ikke en god idé. Nogle gange er det bedre at holde sig for sig selv. Men sådan er det, det er der ikke noget at gøre ved.\n",  # noqa: E501
    },
    {
        "role": "assistant",
        "content": "I extracted the following triplets:\n\tAt blande sig - er ikke - en god idé\n\tdet - er - bedre at holde sig for sig selv\n",  # noqa: E501
    },
    {
        "role": "user",
        "content": "Correct! The next tweet is:\n@fagperson: @Politician1 @minister @journalist Som fagperson vil @fagperson gerne understrege at mine inputs burde fylde mere. \nDet er vigtigt at lytte til hvad @fagperson siger, og I er alle myndighedspersoner der burde vægte sådanne faglige indspark ekstra højt.\n",  # noqa: E501
    },
    {
        "role": "assistant",
        "content": "I extracted the following triplets:\n\t@fagperson - vil - gerne understrege\n\tmine inputs - burde fylde - mere\n\tDet - er vigtigt - at lytte til hvad @fagperson siger\n\tmyndighedspersoner - burde vægte - sådanne faglige indspark ekstra højt\n",  # noqa: E501
    },
    {
        "role": "user",
        "content": "Correct! The next tweet is:\n@minister: @minister lytter til @fagperson og @almindelig123. @minister er enig med dem. Solid argumentation har overbevist mig om at flertallet nok har ret.\n",  # noqa: E501
    },
    {
        "role": "assistant",
        "content": "I extracted the following triplets:\n\t@minister - lytter - til @fagperson og @almindelig123\n\t@minister - er enig - med dem\n\tSolid argumentation - har overbevist - mig om at flertallet nok har ret\n",  # noqa: E501
    },
    {
        "role": "user",
        "content": "The tweet is:\n@user1: @user2 This is a test tweet, I am commenting on something someone else said. @user2 this is not good enough.\n",  # noqa: E501
    },
]



import pytest
from conspiracies.docprocessing.relationextraction.gptprompting import (
    MarkdownPromptTemplate1,
    MarkdownPromptTemplate2,
    PromptTemplate1,
    PromptTemplate2,
    XMLStylePromptTemplate,
    chatGPTPromptTemplate,
)

from .test_data.prompt_data import (
    MarkdownPromptTemplate1_expected_response,
    MarkdownPromptTemplate1_expected_triplets,
    MarkdownPromptTemplate2_expected_response,
    MarkdownPromptTemplate2_expected_triplets,
    PromptTemplate1_expected_response,
    PromptTemplate1_expected_triplets,
    PromptTemplate2_expected_response,
    PromptTemplate2_expected_triplets,
    XMLStylePromptTemplate_expected_response,
    XMLStylePromptTemplate_expected_triplets,
    chatGPTPromptTemplate_expected_response,
    chatGPTPromptTemplate_expected_triplets,
    test_tweet,
)
from .utils import docs_with_triplets  # noqa: F401


@pytest.mark.parametrize(
    "template, target, response, expected_triplets",
    [
        (
            PromptTemplate1,
            test_tweet,
            PromptTemplate1_expected_response,
            PromptTemplate1_expected_triplets,
        ),
        (
            PromptTemplate2,
            test_tweet,
            PromptTemplate2_expected_response,
            PromptTemplate2_expected_triplets,
        ),
        (
            MarkdownPromptTemplate1,
            test_tweet,
            MarkdownPromptTemplate1_expected_response,
            MarkdownPromptTemplate1_expected_triplets,
        ),
        (
            MarkdownPromptTemplate2,
            test_tweet,
            MarkdownPromptTemplate2_expected_response,
            MarkdownPromptTemplate2_expected_triplets,
        ),
        (
            XMLStylePromptTemplate,
            test_tweet,
            XMLStylePromptTemplate_expected_response,
            XMLStylePromptTemplate_expected_triplets,
        ),
        (
            chatGPTPromptTemplate,
            test_tweet,
            chatGPTPromptTemplate_expected_response,
            chatGPTPromptTemplate_expected_triplets,
        ),
    ],
)
def test_PromptTemplate_parse_prompt(
    template,
    target,
    response,
    expected_triplets,
    docs_with_triplets,  # noqa: F811
):
    template_instance = template(examples=docs_with_triplets)
    parsed_triplets = template_instance.parse_prompt(response, target)
    for parsed_triplet, expected_triplet in zip(parsed_triplets, expected_triplets):
        assert parsed_triplet == expected_triplet
    parsed_triplets = template_instance.parse_prompt(response, target)
    assert parsed_triplets == expected_triplets



"""Extracting longest thread from twitter corpus."""

import os
from glob import glob
import ndjson


def ndjson_gen(filepath: str):
    for file in glob(filepath):
        with open(file) as f:
            reader = ndjson.reader(f)

            for post in reader:
                yield post


### Longest thread ####
def longest_thread(tweets: list):
    conversations: dict = {}
    conversation_start: dict = {}
    in_reply_to: dict = {}
    for tweet in tweets:
        conv_id = tweet["conversation_id"]
        if conv_id not in conversations.keys():
            conversations[conv_id] = {"replies": []}

        if "in_reply_to_user_id" in tweet.keys():
            conversations[conv_id]["replies"].append(tweet["id"])

            if "referenced_tweets" not in tweet.keys():
                print("reply but no reference?")
                continue

            for ref in tweet["referenced_tweets"]:
                if ref["type"] == "replied_to":
                    replied_to_id = ref["id"]
                    if replied_to_id not in in_reply_to.keys():
                        in_reply_to[replied_to_id] = [tweet]
                    else:
                        in_reply_to[replied_to_id].append(tweet)
        else:
            conversations[conv_id]["start"] = tweet["id"]
            conversation_start[conv_id] = tweet

    def find_longest_thread(tweet: dict, in_reply_to: dict):
        if tweet["id"] not in in_reply_to.keys():
            return [tweet]

        thread: list = []
        for reply in in_reply_to[tweet["id"]]:
            tmp_thread = find_longest_thread(reply, in_reply_to)
            if len(tmp_thread) > len(thread):
                thread = tmp_thread

        thread.insert(0, tweet)
        return thread

    no_reply_tweets = 0
    threads: dict = {}
    for conv_id, conv_tweets in conversations.items():
        if "start" not in conv_tweets.keys():
            continue

        threads[conv_id] = find_longest_thread(conversation_start[conv_id], in_reply_to)

        if len(threads[conv_id]) > 1:
            print("\n\n THREAD STARTS: \n")
            print(f"Conversation id: {conv_id}")
            for t in threads[conv_id]:
                print("\n-----------\n")
                print(t["text"])
            print("\n-----------\n")
        else:
            no_reply_tweets += 1

    print(f"{no_reply_tweets} tweets with no reply out of {len(threads)}")


### context window thread ###
def context_window_thread(tweets):
    conversations: dict = {}
    for tweet in tweets:
        conv_id = tweet["conversation_id"]
        if conv_id not in conversations.keys():
            conversations[conv_id] = {tweet["id"]: tweet}

        else:
            conversations[conv_id][tweet["id"]] = tweet

    def find_reply_to_id(tweet: dict):
        if "referenced_tweets" not in tweet.keys():
            Warning("reply but no reference?")
            return None

        for ref in tweet["referenced_tweets"]:
            if ref["type"] == "replied_to":
                return ref["id"]

    def find_context(tweet: dict, conversation: dict, context_length=4):
        if "in_reply_to_user_id" in tweet.keys():
            if context_length == 0:
                return [tweet]
            else:
                if "referenced_tweets" not in tweet.keys():
                    Warning("reply but no reference?")
                    return [tweet]

                reply_to_id = find_reply_to_id(tweet)
                try:
                    reply_to = conversation[reply_to_id]
                except KeyError:
                    return [tweet]

                context = find_context(
                    reply_to,
                    conversation,
                    context_length=context_length - 1,
                )
                context.append(tweet)
                return context
        else:
            return [tweet]

    for tweet in tweets:
        context = find_context(tweet, conversations[tweet["conversation_id"]])

        if len(context) > 1:
            print(f'TWEET ID: {tweet["id"]}')
            for context_tweet in context:
                print(context_tweet["text"])
                print("\n-----------\n")


if __name__ == "__main__":
    path = os.path.join("/data", "004_twitter-stopword", "*.ndjson")

    tweets = []
    for i, post in enumerate(ndjson_gen(path)):
        tweets.append(post)
        if i == 1000:
            break

    context_window_thread(tweets)



from typing import List

import pytest
from confection import registry
from conspiracies.docprocessing.relationextraction.gptprompting import (
    SpanTriplet,
    StringTriplet,
)
from spacy.language import Language

from .test_prompt_template_parse_prompt import (
    MarkdownPromptTemplate1_expected_response,
    MarkdownPromptTemplate1_expected_triplets,
)
import spacy
from .utils import docs_with_triplets, nlp_da  # noqa F401

thread1 = """
@user1: This is a test tweet, I am commenting on something someone else said. END
@user1: @user2 this is not good enough END
    """

sample_threads = [thread1]
api_responses = [MarkdownPromptTemplate1_expected_response]
expected_triplets = [MarkdownPromptTemplate1_expected_triplets]


@pytest.mark.parametrize(
    "sample_thread, api_response, expected_triplets",
    list(zip(sample_threads, api_responses, expected_triplets)),
)
def test_prompt_relation_extraction(
    nlp_da: Language,  # noqa F811
    sample_thread: str,
    api_response: str,
    expected_triplets: List[StringTriplet],
):
    nlp = nlp_da
    expected_span_triplets = [
        SpanTriplet.from_doc(triplet, nlp(sample_thread))
        for triplet in expected_triplets
    ]

    @registry.prompt_apis.register("test_api")  # type: ignore
    def create_test_api(prompt_template, api_key, model_name, api_kwargs):
        """A test api for testing purposes.

        assumes prompttemplate is markdown_template_2
        """

        def test_api(target):
            return [api_response]

        return test_api

    # test that component can be added.
    config = {
        "prompt_template": "conspiracies/markdown_template_2",
        "examples": None,
        "task_description": None,
        "model_name": "text-davinci-002",
        "backend": "test_api",
        "api_key": "empty-key",
        "api_kwargs": {
            "max_tokens": 500,
            "temperature": 0.7,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
        },
        "force": True,
    }
    nlp.add_pipe("conspiracies/prompt_relation_extraction", config=config)

    doc = nlp(sample_thread)

    doc_relation_triplets = doc._.relation_triplets
    assert doc_relation_triplets is not None and len(doc_relation_triplets) > 0
    for triplet in doc_relation_triplets:
        assert isinstance(triplet, SpanTriplet)
        triplet in expected_span_triplets


def test_create_pipeline():
    nlp = spacy.blank("da")
    config = {"api_key": ""}

    nlp.add_pipe("conspiracies/prompt_relation_extraction", config=config)

    # remove pipeline
    nlp.remove_pipe("conspiracies/prompt_relation_extraction")

    # add pipeline with speicifc config
    config = {
        "prompt_template": "conspiracies/template_1",
        "examples": None,
        "task_description": None,
        "model_name": "text-davinci-002",
        "backend": "conspiracies/openai_gpt3_api",
        "api_key": "",
        "split_doc_fn": None,
        "api_kwargs": {
            "max_tokens": 500,
            "temperature": 0.7,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
        },
        "force": True,
    }

    nlp.add_pipe(
        "conspiracies/prompt_relation_extraction",
        last=True,
        config=config,
    )



"""Headwords extraction as a spaCy component."""

from collections import Counter
from typing import Union
from warnings import warn

from spacy.language import Language
from spacy.tokens import Doc, Span, Token


class HeadwordsExtractionComponent:
    """A class for extracting headwords from a given text document.

    Args:
        nlp (Language): Language processing pipelines.
        name (str): The instance name of the component in the pipeline.

    Attributes:
        raise_error (bool): If True, raises warning message in case no ancestor is found
            within a span.
        normalize_to_entity (bool): If True, normalizes a token to an entity.
        normalize_to_noun_chunk (bool): If True, normalizes a token to a noun chunk.
    """

    def __init__(
        self,
        nlp: Language,
        name: str,
        raise_error: bool,
        normalize_to_entity: bool,
        normalize_to_noun_chunk: bool,
        force: bool,
    ):
        """Initialise components."""

        self.raise_error = raise_error
        self.normalize_to_entity = normalize_to_entity
        self.normalize_to_noun_chunk = normalize_to_noun_chunk
        self.name = name

        if not Token.has_extension("to_span") or force:
            Token.set_extension("to_span", getter=self.to_span, force=force)
        if not Span.has_extension("to_span") or force:
            Span.set_extension("to_span", getter=lambda span: span, force=force)
        if not Doc.has_extension("to_span") or force:
            Doc.set_extension("to_span", getter=lambda doc: doc[:], force=force)

        if (not Doc.has_extension("most_common_ancestor")) or force:
            Doc.set_extension(
                "most_common_ancestor",
                getter=lambda doc: self.most_common_ancestor(doc[:]),
                force=force,
            )
        if (not Span.has_extension("most_common_ancestor")) or force:
            Span.set_extension(
                "most_common_ancestor",
                getter=self.most_common_ancestor,
                force=force,
            )

    def to_entity(self, token: Token) -> Span:  # type: ignore
        """Normalize token to an entity.

        Args:
            token(Token): The token to normalize.

        Returns:
            Span: The entity.
        """
        for ent in token.doc.ents:
            if token in ent:
                return ent

    def to_noun_chunk(self, token: Token) -> Span:  # type: ignore
        """Normalize token to a noun chunk.

        Args:
            token(Token): The token to normalize.

        Returns:
            Span: The noun chunk.
        """
        doc = token.doc
        for noun_chunk in doc.noun_chunks:
            if token in noun_chunk:
                return noun_chunk

    def to_span(self, token: Union[Token, Span, Doc]) -> Span:
        """Normalize token to a span.

        Args:
            token(Token): The token to normalize.

        Returns:
            Span: The normalized token.
        """
        if self.normalize_to_entity:
            ent = self.to_entity(token)  # type: ignore
            if ent:
                return ent

        if self.normalize_to_noun_chunk:
            noun_chunk = self.to_noun_chunk(token)  # type: ignore
            if noun_chunk:
                return noun_chunk

        doc = token.doc
        return doc[token.i : token.i + 1]  # type: ignore

    def most_common_ancestor(
        self,
        span: Union[Doc, Span],
        raise_error: bool = False,
    ) -> Span:
        """Find the most common ancestor in a span.

        Args:
           span(Span): The span to find the most common ancestor of.
           raise_error(bool): Raises warning message if no ancestor is found within a
            span.

        Returns:
            Span: The most common ancestor of the span.
        """
        ancestors_in_span = Counter(
            [
                ancestor
                for token in span
                for ancestor in token.ancestors
                if ancestor in span
            ],
        )
        most_common_ancestors = ancestors_in_span.most_common()
        if most_common_ancestors:
            most_common_ancestor = most_common_ancestors[0][0]
        else:  # fall back is to simply take the first token
            most_common_ancestor = span[0]

        normalized_token = self.to_span(most_common_ancestor)

        if len(normalized_token) != 1:
            error_message = (
                f"None of the tokens in the span ({span}) contains an"
                + " ancestor within this span."
            )

            if raise_error:
                warn(error_message)

        return normalized_token

    def __call__(self, doc: Doc):
        """Run the pipeline component."""
        return doc


def contains_ents(span: Union[Doc, Span]) -> bool:
    """Check if a Doc or span contains entities.

    Args:
        span (Union[Doc, Span]): The span to find entites in.

    Returns:
        bool: Returns True if a token has an entity label.
    """
    for token in span:
        if token.ent_type:
            return True
    return False


@Language.factory(
    "heads_extraction",
    default_config={
        "raise_error": False,
        "normalize_to_entity": True,
        "normalize_to_noun_chunk": True,
        "force": True,
    },
)
def create_headwords_component(
    nlp: Language,
    name: str,
    raise_error: bool,
    normalize_to_entity: bool,
    normalize_to_noun_chunk: bool,
    force: bool,
) -> HeadwordsExtractionComponent:
    """Allows HeadwordsExtraction to be added to a spaCy pipe using
    nlp.add_pipe("heads_extraction").

    Args:
        nlp (Language): Language processing pipelines.

        name (str): The instance name of the component in the pipeline.
        raise_error (bool): If True, raises warning message in case no ancestor is found
            within a span.
        normalize_to_entity (bool): If True, normalizes a token to an entity.
        normalize_to_noun_chunk (bool): If True, normalizes a token to a noun chunk.

    Returns:
        HeadwordsExtraction: A spaCy component.
    """

    return HeadwordsExtractionComponent(
        nlp,
        name=name,
        raise_error=raise_error,
        normalize_to_entity=normalize_to_entity,
        normalize_to_noun_chunk=normalize_to_noun_chunk,
        force=force,
    )



import os.path
from typing import Iterable

import spacy
from jsonlines import jsonlines
from tqdm import tqdm

from conspiracies import docs_to_jsonl
from conspiracies.document import Document, text_with_context, remove_context


class DocProcessor:
    def _build_coref_pipeline(self):
        nlp_coref = spacy.blank("da")
        nlp_coref.add_pipe("sentencizer")
        nlp_coref.add_pipe("allennlp_coref")

        def warn_error(proc_name, proc, docs, e):
            print(
                f"An error occurred when applying component {proc_name}. "
                f"{[d[:20] + '...' for d in docs]}. {e}",
            )

        nlp_coref.set_error_handler(warn_error)
        return nlp_coref

    def _build_triplet_extraction_pipeline(self):
        nlp = spacy.load("da_core_news_sm")
        nlp.add_pipe(
            "heads_extraction",
            config={"normalize_to_entity": True, "normalize_to_noun_chunk": True},
        )
        # TODO: Any and only (!) relevant configuration from pipeline config should
        #  trickle down to configuration of individual components here. For now, this
        #  is just copy-pasta from elsewhere.
        if self.triplet_extraction_component.lower() == "multi2oie":
            config = {"confidence_threshold": 2.7, "model_args": {"batch_size": 10}}
            nlp.add_pipe("relation_extractor", config=config)
        elif self.triplet_extraction_component.lower() == "prompting":
            config = {
                "prompt_template": "conspiracies/template_1",
                "examples": None,
                "task_description": None,
                "model_name": "text-davinci-002",
                "backend": "conspiracies/openai_gpt3_api",
                "api_key": "",
                "split_doc_fn": None,
                "api_kwargs": {
                    "max_tokens": 500,
                    "temperature": 0.7,
                    "top_p": 1,
                    "frequency_penalty": 0,
                    "presence_penalty": 0,
                },
                "force": True,
            }

            nlp.add_pipe(
                "conspiracies/prompt_relation_extraction",
                last=True,
                config=config,
            )
        else:
            ValueError(
                f"{self.triplet_extraction_component} is not a valid triplet "
                f"relation extraction component.",
            )
        return nlp

    def __init__(self, triplet_extraction="multi2oie"):
        self.coref_pipeline = self._build_coref_pipeline()
        self.triplet_extraction_component = triplet_extraction
        self.triplet_extraction_pipeline = self._build_triplet_extraction_pipeline()

    def process_docs(
        self,
        docs: Iterable[Document],
        output_path: str,
        continue_from_last=False,
        batch_size=25,
    ):
        if continue_from_last and os.path.exists(output_path):
            with jsonlines.open(output_path) as annotated_docs:
                already_processed = {
                    annotated_doc["id"] for annotated_doc in annotated_docs
                }
            print(f"Skipping {len(already_processed)} processed docs.")
            docs = (doc for doc in docs if doc["id"] not in already_processed)

        # The coreference pipeline tends to choke on too large batches because of an
        # extreme memory pressure, hence the small batch size
        coref_resolved_docs = self.coref_pipeline.pipe(
            ((text_with_context(doc), doc["id"]) for doc in docs),
            batch_size=batch_size,
            as_tuples=True,
        )

        with_triplets = self.triplet_extraction_pipeline.pipe(
            (
                (remove_context(doc._.resolve_coref), id_)
                for doc, id_ in coref_resolved_docs
            ),
            batch_size=batch_size * 4,
            as_tuples=True,
        )

        docs_to_jsonl(
            tqdm(d for d in with_triplets),
            output_path,
            append=continue_from_last,
        )



"""Extracting context threads from twitter corpus."""

import os
from glob import glob
import ndjson
import argparse
import datetime
import re


def ndjson_gen(filepath: str):
    """Creates a generator for ndjson files and yields the tweets without
    retweets."""
    for file in glob(filepath):
        print("Reading file:", file)
        with open(file) as f:
            reader = ndjson.reader(f)
            for post in reader:
                if re.search("^RT", post["text"]):  # remove retweets
                    continue
                yield post


def context_window_thread(tweets, context_len):
    """
    Extracts a thread of context tweets; the last tweet in the thread is the one that the context relates to
    Args:
        tweets: list of tweets
        context_len: number of tweets to include in context
    Returns:
        contexts: list of contexts, where each context is a list of tweets
    """
    conversations = {}
    for tweet in tweets:
        conv_id = tweet["conversation_id"]
        # create a new conversation if it doesn't exist
        conversations.setdefault(conv_id, {})
        conversations[conv_id][tweet["id"]] = tweet

    def find_context(tweet, conversation, context_length=context_len):
        # if the tweet is the first tweet in the conversation or the context length is 0, return the tweet
        if (
            "in_reply_to_user_id" not in tweet
            or context_length == 0
            or "referenced_tweets" not in tweet
        ):
            return [tweet]
        # find the tweet that this tweet is replying to
        replied_to_refs = [
            ref for ref in tweet["referenced_tweets"] if ref["type"] == "replied_to"
        ]
        # if the tweet is not replying to another tweet, return the tweet
        if not replied_to_refs:
            return [tweet]
        # if the tweet is replying to another tweet, find the context for that tweet
        reply_to_id = replied_to_refs[0]["id"]
        # if the tweet is replying to a tweet that is not in the conversation, return the tweet
        if reply_to_id not in conversation:
            return [tweet]
        # find the context for the tweet that this tweet is replying to and add the tweet to the context
        context = find_context(
            conversation[reply_to_id],
            conversation,
            context_length=context_length - 1,
        )
        return context + [tweet]

    # find the context for each tweet in the conversation and return the contexts that are longer than 1 tweet
    contexts = [
        find_context(tweet, conversations[tweet["conversation_id"]]) for tweet in tweets
    ]
    return contexts


def range_of_dates(start_date, extra_days):
    start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d").date()
    date_range = [start_date + datetime.timedelta(days=x) for x in range(extra_days)]
    return date_range


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Get tweets + context from date range")
    parser.add_argument(
        "-s",
        "--start_date",
        type=str,
        default="2019-05-22",
        help="start date in format YYYY-MM-DD",
    )
    parser.add_argument(
        "-e",
        "--extra_days",
        type=int,
        default=15,
        help="number of days to include from start date and after",
    )
    parser.add_argument(
        "-p",
        "--data_path",
        type=str,
        default="da_stopwords_part?_2019-01-01_2020-12-31.ndjson",
        help='path to ndjson file; for all: "*.ndjson"',
    )
    parser.add_argument(
        "-c",
        "--context_len",
        type=int,
        default=4,
        help="number of tweets to include in context",
    )
    args = parser.parse_args()

    # path to twitter data
    path = os.path.join("/data", "004_twitter-stopword", args.data_path)

    date_range = range_of_dates(args.start_date, args.extra_days)
    print(date_range)
    print(
        f"Looking for tweets from {args.start_date} and {args.extra_days} days forward (to and indcluding: {datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() + datetime.timedelta(days=args.extra_days-1)})",
    )
    tweets = []

    for i, post in enumerate(ndjson_gen(path)):
        # get the date of the tweet and append the tweet if it is within the date range
        date_ = datetime.datetime.strptime(post["created_at"][0:10], "%Y-%m-%d").date()
        if date_ in date_range:
            tweets.append(post)
        # if len(tweets) > 1000:
        #     break
    print(f"Found {len(tweets)} tweets in the date range (out of {i} tweets total)")

    # run the tweet thread extraction
    contexts = context_window_thread(tweets, args.context_len)

    # write contests to ndjson; the ndjson of contexts has the format: List[Dict[str, Any]]
    with open(
        os.path.join(
            f"tweet_threads_{datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() }_{datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() + datetime.timedelta(days=args.extra_days-1)}.ndjson",
        ),
        "w",
    ) as f:
        ndjson.dump(contexts, f)



from typing import Any

import toml
from pydantic import BaseModel


class BaseConfig(BaseModel):
    project_name: str
    output_root: str
    language: str


class StepConfig(BaseModel):
    enabled: bool = True


class PreProcessingConfig(StepConfig):
    input_path: str
    doc_type: str
    metadata_fields: set[str]
    extra: dict = {}


class DocProcessingConfig(StepConfig):
    triplet_extraction_method: str


class CorpusProcessingConfig(StepConfig):
    pass


class PipelineConfig(BaseModel):
    base: BaseConfig
    preprocessing: PreProcessingConfig
    docprocessing: DocProcessingConfig
    corpusprocessing: CorpusProcessingConfig

    @staticmethod
    def update_nested_dict(d: dict[str, Any], path: str, value: Any) -> None:
        """Sets or updates a nested dict in place with a TOML-like string key.

        >>> some_dict = {"a": {"b": 1}}
        >>> PipelineConfig.update_nested_dict(some_dict, "a.b", 2)
        >>> some_dict["a"]["b"] == 2  # True

        Args:
            d: (nested) dict with str keys to be updated
            path: a TOML-like string key with '.' as level separators, e.g. "a.b.c"
            value: value to set under the nested key
        """
        keys = path.split(".")
        for key in keys[:-1]:
            d = d.setdefault(key, {})
        d[keys[-1]] = value

    @classmethod
    def from_toml_file(cls, path: str, extra_config: dict = None):
        with open(path, "r") as file:
            config_data = toml.load(file)
        if extra_config:
            for path, value in extra_config.items():
                if value is None:
                    continue
                PipelineConfig.update_nested_dict(config_data, path, value)
        return cls(**config_data)



"""A custom Coreference model for wrapping an AllenNLP's Predictor for
coreference resolution on SpaCy Docs."""

from typing import List, Union
from pathlib import Path

from spacy.tokens import Doc

from allennlp.predictors.predictor import Predictor
from allennlp.data import Instance
from allennlp.models.archival import load_archive
from allennlp.common.util import prepare_environment

# required for reading an archive using the "coref"
from allennlp_models.coref.dataset_readers.conll import ConllCorefReader  # noqa

from conspiracies.docprocessing.modeldownload import download_model


@Predictor.register("coreference_resolution_v1")
class CoreferenceModel(Predictor):
    """A coreference predictor model using the AllenNLP API, but for handling
    SpaCy docs.

    Args:
        model_path(Union[Path, str, None], optional): Path to the model, if None, the
            model will be downloaded to the default cache directory.
        device(int, optional): Cuda device. If >= 0 will use the corresponding GPU,
            below 0 is CPU. Defaults to -1.
        open_unverified_connection (bool, optional): Should you download from an
            unverified connection. Defaults to False.

    Returns:
        CoreferenceModel: The coreference model
    """

    def __init__(
        self,
        model_path: Union[Path, str, None] = None,
        device: int = -1,
        open_unverified_connection: bool = False,
        **kwargs,
    ) -> None:
        if model_path is None:
            model_path = download_model(
                "da_coref_twitter_v1",
                open_unverified_connection=open_unverified_connection,
            )

        archive = load_archive(model_path, cuda_device=device, **kwargs)
        config = archive.config
        prepare_environment(config)
        dataset_reader = archive.validation_dataset_reader
        super().__init__(archive.model, dataset_reader)

    def _doc_to_instance(self, doc: Doc) -> Instance:
        sentences = [[token.text for token in sentence] for sentence in doc.sents]
        instance = self._dataset_reader.text_to_instance(sentences)
        return instance

    def predict_batch_docs(self, docs: List[Doc]) -> List[Instance]:
        """Convert a list of docs to Instance and predict the batch."""
        instances = [self._doc_to_instance(doc) for doc in docs]
        return self.predict_batch_instance(instances)



from pathlib import Path
from typing import List, Optional

import spacy
from conspiracies import docs_from_jsonl
from spacy.tokens import Doc


def load_gold_triplets(
    path: Optional[str] = None,
    nlp: Optional[spacy.Language] = None,
) -> List[Doc]:
    if nlp is None:
        nlp = spacy.blank("en")
        nlp.add_pipe("sentencizer")
    if path is None:
        path = "/data/conspiracies/gold_triplets/gold_triplets_tweet.jsonl"
    # check if it existing
    if not Path(path).exists():
        raise FileNotFoundError(
            f"File {path} not found. You are probably not running this from Grundtvig"
            + ", in which case you will have to specify the path.",
        )
    return docs_from_jsonl(path, nlp)


def load_api_key() -> str:
    path = Path("/data") / "conspiracies" / "api_key.txt"
    if not path.exists():
        raise FileNotFoundError(
            f"File {path} not found. You are probably not running this from Grundtvig"
            + ", in which case you will have to specify the path.",
        )
    with open(path, "r") as f:
        key = f.read()
    return key



import json

import bs4

from conspiracies.common.fileutils import iter_lines_of_files
from conspiracies.document import Document
from conspiracies.preprocessing.preprocessor import Preprocessor


class InfoMediaPreprocessor(Preprocessor):
    METADATA_KEYS = {
        "ArticleUrl",
        "Authors",
        "Heading",
        "Lead",
        "PageIds",
        "PublishDate",
        "Section",
        "Source",
        "WordCount",
    }

    def _do_preprocess_docs(self, glob_pattern: str):
        lines = iter_lines_of_files(glob_pattern)
        for line in lines:
            yield self.process_line(line)

    def process_line(self, line: str):
        obj = json.loads(line)

        doc_id = obj["ArticleId"]
        metadata = {k: obj[k] for k in InfoMediaPreprocessor.METADATA_KEYS}
        text = self.create_text(obj)

        return Document(id=doc_id, metadata=metadata, text=text, context=None)

    @staticmethod
    def create_text(doc_obj: dict):
        text_pieces = [doc_obj["Heading"] + "\n"]

        subheading = doc_obj["SubHeading"]
        if subheading:
            text_pieces.append(subheading + "\n")

        body = bs4.BeautifulSoup(doc_obj["BodyText"], features="html.parser")
        paragraphs = body.find_all("p")
        for paragraph in paragraphs:
            paragraph_pieces = []
            for child in paragraph:
                if child.name == "br":
                    paragraph_pieces.append("\n")
                else:
                    paragraph_pieces.append(child.get_text())
            text_pieces.append("".join(paragraph_pieces))
        return "\n".join(text_pieces).strip()



import conspiracies  # noqa F401

from .utils import nlp_en  # noqa F401


def test_most_common_ancestor(nlp_en):  # noqa F811
    nlp_en.add_pipe(
        "heads_extraction",
        config={"normalize_to_entity": False, "normalize_to_noun_chunk": False},
    )
    doc = nlp_en("Mette Frederiksen is the Danish politician.")

    assert doc[0:1]._.most_common_ancestor.text == "Mette"  # Single token Span
    assert doc[0:2]._.most_common_ancestor.text == "Frederiksen"  # Span
    assert doc._.most_common_ancestor.text == "is"  # Doc

    # remove the pipe
    nlp_en.remove_pipe("heads_extraction")
    nlp_en.add_pipe(
        "heads_extraction",
        config={
            "force": True,
            "normalize_to_entity": True,
            "normalize_to_noun_chunk": True,
        },
    )

    assert (
        doc[0:1]._.most_common_ancestor.text == "Mette Frederiksen"
    )  # Single token Span
    assert doc[0:2]._.most_common_ancestor.text == "Mette Frederiksen"  # Span
    assert doc._.most_common_ancestor.text == "is"  # Doc



"""Dataset class used for preparing input to relation extraction model."""

import torch
from torch.utils.data import Dataset

from .other import utils
from .multi2oie_utils import get_cached_tokenizer


class EvalDataset(Dataset):
    def __init__(
        self,
        data_path,
        max_len,
        from_list=True,
        tokenizer_config="bert-base-multilingual-cased",
    ):
        if not from_list:
            self.sentences = utils.load_pkl(data_path)
        else:
            # Data should be a list of sentences (or pickled list of sentences)
            self.sentences = data_path
        self.tokenizer = get_cached_tokenizer(tokenizer_config)
        self.vocab = self.tokenizer.vocab
        self.max_len = max_len

        self.pad_idx = self.vocab["[PAD]"]
        self.cls_idx = self.vocab["[CLS]"]
        self.sep_idx = self.vocab["[SEP]"]
        self.mask_idx = self.vocab["[MASK]"]

    def add_pad(self, token_ids):
        diff = self.max_len - len(token_ids)
        if diff > 0:
            token_ids += [self.pad_idx] * diff
        else:
            token_ids = token_ids[: self.max_len - 1] + [self.sep_idx]
        return token_ids

    def idx2mask(self, token_ids):
        return [token_id != self.pad_idx for token_id in token_ids]

    def __getitem__(self, idx):
        token_ids = self.add_pad(self.tokenizer.encode(self.sentences[idx]))
        att_mask = self.idx2mask(token_ids)
        token_strs = self.tokenizer.convert_ids_to_tokens(token_ids)
        sentence = self.sentences[idx]

        assert len(token_ids) == self.max_len
        assert len(att_mask) == self.max_len
        assert len(token_strs) == self.max_len
        batch = [torch.tensor(token_ids), torch.tensor(att_mask), token_strs, sentence]
        return batch

    def __len__(self):
        return len(self.sentences)



from data import load_api_key
import spacy
from data import load_gold_triplets
from extract_examples import extract_examples
from spacy.training import Example


def save_string(path, text_string, method="w"):
    with open(path, method) as f:
        f.write(text_string)


def create_pipeline(
    template: str,
    n_target: int = 10,
    cv: int = 2,
):
    nlp = spacy.load("da_core_news_trf")
    gold_docs = load_gold_triplets(nlp=nlp)
    targets, examples = extract_examples(gold_docs, n_target, cv)
    if template == "chatgpt_style_template":
        model_name = "gpt-3.5-turbo"
        backend = "conspiracies/openai_chatgpt_api"
    else:
        model_name = "text-davinci-002"
        backend = "conspiracies/openai_gpt3_api"
    config = {
        "prompt_template": f"conspiracies/{template}",
        "examples": None,
        "task_description": None,
        "model_name": model_name,
        "backend": backend,
        "api_key": load_api_key(),
        "split_doc_fn": None,
        "api_kwargs": {
            "max_tokens": 500,
            "temperature": 0.7,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
        },
        "force": True,
    }
    relation_component = nlp.add_pipe(
        "conspiracies/prompt_relation_extraction",
        last=True,
        config=config,
    )

    for cv, (target_set, example_set) in enumerate(zip(targets, examples)):
        prompt_template = relation_component.prompt_template
        prompt_template.examples = example_set
        docs = nlp.pipe([doc.text for doc in target_set])
        docs = list(docs)
        predictions = [
            Example(predicted=pred_doc, reference=gold_doc)
            for pred_doc, gold_doc in zip(docs, target_set)
        ]
        scores = relation_component.score(predictions)
        print(f"Scores for round {cv} using template {template}:")
        scores.pop("sample_scores")
        print(scores)
        row = f'|{template} {cv+1} | {scores["exact_span_match_f1"]} | {scores["exact_string_match_f1"]} | {scores["normalized_span_overlap_f1"]} | {scores["normalized_string_overlap_f1"]} | {scores["n_predictions"]} | {scores["n_references"]} |\n'
        save_string("results.md", row, "a")


templates = [
    # "template_1",
    # "template_2",
    # "markdown_template_1",
    # "markdown_template_2",
    # "xml_style_template",
    "chatgpt_style_template",
]
header = """
Socring: 
|Template CV | exact span match f1 | exact string match f1 | norm span overlap f1 | norm string overlap f1 | n predictions | n references |
| --- | --- | --- | --- | --- | --- | --- |
"""

# save_string("results.md", header)
for template in templates:
    create_pipeline(template)



from typing import List

from catalogue import registry
from spacy.tokens import Doc, Span


@registry.split_doc_functions.register("conspiracies/split_thread_to_tweets")
def split_thread_to_tweets(doc: Doc) -> List[Span]:
    """Split a thread into tweets.

    Args:
        doc (Doc): A spaCy Doc object.

    Returns:
        List[Doc]: A list of spaCy Doc objects, each representing a tweet.
    """
    # split doc by newline
    tweets = []

    start = 0
    end = 0
    for token in doc:
        if token.text == "\n":  # TODO: check that this actually splits the thread
            end = token.i
            tweets.append(doc[start:end])
            start = end + 1
            tweets.append(doc[start:])
    return tweets



"""A SpaCy component for coference using an AllenNLP coference model."""

from pathlib import Path
from typing import Dict, Iterable, Iterator, Union

from spacy import Vocab
from spacy.language import Language
from spacy.pipeline import TrainablePipe
from spacy.tokens import Doc, Span
from spacy.util import minibatch

from conspiracies.docprocessing.coref import CoreferenceModel


class CoreferenceComponent(TrainablePipe):
    def __init__(
        self,
        vocab: Vocab,
        name: str,
        model_path: Union[Path, str, None],
        device: int,
        open_unverified_connection: bool,
    ):
        self.name = name
        self.vocab = vocab
        self.model = CoreferenceModel(  # type: ignore
            model_path=model_path,
            device=device,
            open_unverified_connection=open_unverified_connection,
        )

        # Register custom extension on the Doc and Span
        if not Doc.has_extension("resolve_coref"):
            Doc.set_extension("resolve_coref", getter=self.resolve_coref_doc)
        if not Span.has_extension("resolve_coref"):
            Span.set_extension("resolve_coref", getter=self.resolve_coref_span)
        if not Doc.has_extension("coref_clusters"):
            Doc.set_extension("coref_clusters", default=list())
        if not Span.has_extension("coref_clusters"):
            Span.set_extension("coref_clusters", default=list())
        if not Span.has_extension("antecedent"):
            Span.set_extension("antecedent", default=None)

    def resolve_coref_doc(self, doc: Doc) -> str:
        """Resolve the coreference clusters by replacing each entity with the
        antecedent. The antecedent is the first entity that appears in the
        cluster. This is for the whole doc.

        Args:
            doc (Doc): The document.

        Returns:
            Str: The text of the document with resolved coreference clusters.
        """
        resolved = list(tok.text_with_ws for tok in doc)
        for i, cluster in doc._.coref_clusters:
            for coref in cluster:
                if coref != coref._.antecedent:
                    resolved[coref.start] = (
                        coref._.antecedent.text + doc[coref.end - 1].whitespace_
                    )
                    for i in range(coref.start + 1, coref.end):
                        resolved[i] = ""
        return "".join(resolved)

    def resolve_coref_span(self, sent: Span) -> str:
        """Resolve the coreference clusters by replacing each entity with the
        antecedent. The antecedent is the first entity that appears in the
        cluster. This is for the the sent.

        Args:
            sent (Span): The document.

        Returns:
            Str: The text of the sentence with resolved coreference clusters.
        """
        resolved_span = list(tok.text_with_ws for tok in sent)

        # Calibrate coref index since it is based on the Doc level
        sentence_lengths = [len(sent) for sent in sent.doc.sents]
        index_calibrate = 0
        for i, doc_sent in enumerate(sent.doc.sents):
            if doc_sent == sent:
                break
            else:
                index_calibrate += sentence_lengths[i]

        for i, coref in sent._.coref_clusters:
            coref_start = coref.start - index_calibrate
            coref_end = coref.end - index_calibrate
            if coref != coref._.antecedent:
                resolved_span[coref_start] = (
                    coref._.antecedent.text + sent.doc[coref_end].whitespace_
                )
                for i in range(coref_start + 1, coref_end):
                    resolved_span[i] = ""
        return "".join(resolved_span).strip()

    def set_annotations(self, docs: Iterable[Doc], model_output) -> None:
        """Set the coref attributes on Doc and Token level
        Args:
            docs (Iterable[Doc]): The documents to modify.
            model_output: (Dict): A batch outåut of self.predict().
        """
        for doc, prediction in zip(docs, model_output):
            clusters = prediction["clusters"]
            coref_clusters = [
                (
                    clusters.index(d),
                    [doc[cluster_ids[0] : cluster_ids[1] + 1] for cluster_ids in d],
                )
                for d in clusters
            ]
            doc._.coref_clusters.append(coref_clusters)
            doc._.coref_clusters = coref_clusters
            coref_clusters_lookup_dict = dict(coref_clusters)
            for sent in doc.sents:
                for cluster, corefs in coref_clusters:
                    for coref in corefs:
                        coref._.antecedent = corefs[0]
                        if sent == coref.sent:
                            sent._.coref_clusters.append((cluster, coref))
                            sent._.antecedent = coref_clusters_lookup_dict[cluster][0]

    def pipe(self, stream: Iterable[Doc], *, batch_size: int = 128) -> Iterator[Doc]:
        """Apply the pipe to a stream of documents. This usually happens under
        the hood when the nlp object is called on a text and all components are
        applied to the Doc. Batch size is controlled by `batch_size` when
        instatiating the nlp.pipe object.

        Args:
            stream (Iterable[Doc]): A stream of documents.
            batch_size (int): The number of documents to buffer.

        Yields:
            Doc: Processed documents in order.
        """
        for outer_batch in minibatch(stream, batch_size):
            outer_batch = list(outer_batch)
            self.set_annotations(outer_batch, self.predict(outer_batch))

            yield from outer_batch

    def predict(self, docs: Iterable[Doc]) -> Dict:
        """Apply the pipeline's model to a batch of docs, without modifying
        them. Returns the extracted features as the FullTransformerBatch
        dataclass.

        Args:
            docs (Iterable[Doc]): The documents to predict.

        Returns:
            Dict: The extracted features.
        """
        return self.model.predict_batch_docs(docs)

    def __call__(self, doc: Doc) -> Doc:
        """Apply the pipe to one document. The document is modified in place,
        and returned. This usually happens under the hood when the nlp object
        is called on a text and all components are applied to the Doc.

        Args:
            doc (Doc): The Doc to process.

        Returns:
            Doc: The processed Doc.
        """
        outputs = self.predict([doc])
        self.set_annotations([doc], outputs)
        return doc


@Language.factory(
    "allennlp_coref",
    default_config={
        "model_path": None,
        "device": -1,
        "open_unverified_connection": True,
    },
)
def create_coref_component(
    nlp: Language,
    name: str,
    model_path: Union[Path, str, None],
    device: int,
    open_unverified_connection: bool,
):
    """Creates coference model component.

    Args:
        nlp (Language): A spacy language pipeline
        name (str): The name of the component
        model_path (Union[Path, str, None]): Path to the model, if None, the
            model will be downloaded to the default cache directory.
        device (int, optional): Cuda device. If >= 0 will use the corresponding GPU,
            below 0 is CPU. Defaults to -1.
        open_unverified_connection (bool, optional): Should you download the model from
            an unverified connection. Defaults to True.

    Returns:
        CorefenceComponent: The coreference model component
    """

    return CoreferenceComponent(
        nlp.vocab,
        name=name,
        model_path=model_path,
        device=device,
        open_unverified_connection=open_unverified_connection,
    )



import pytest

from .utils import nlp_da  # noqa

from conspiracies.preprocessing.wordpiece_length_normalization import (
    wordpiece_length_normalization,
)
from transformers import AutoTokenizer


@pytest.fixture()
def tokenizer():
    return AutoTokenizer.from_pretrained("alexandrainst/da-sentiment-base")


def test_wp_length_normalization_short(nlp_da, tokenizer):  # noqa
    short_texts = ["hello there"]
    norm_text = wordpiece_length_normalization(short_texts, nlp_da, tokenizer)
    norm_text = list(norm_text)

    assert len(norm_text) == 1, "a short text should only result of one normalized text"
    assert norm_text[0] == short_texts[0], "a short text should result the same text"


def test_wp_length_normalization_long(nlp_da, tokenizer):  # noqa
    long_text = ["Hej mit navn er Kenneth. " * 200]
    norm_text = wordpiece_length_normalization(long_text, nlp_da, tokenizer)
    norm_text = list(norm_text)

    assert len(norm_text) > 1, "a long text should be split into multiple texts"
    assert norm_text[0] == norm_text[1], "splits are not consistent"


def test_wp_length_normalization_no_sent(nlp_da, tokenizer):  # noqa
    # irregular example without sentences
    irregular_text = ["hello " * 600]
    norm_text = wordpiece_length_normalization(irregular_text, nlp_da, tokenizer)
    norm_text = list(norm_text)

    assert (
        len(norm_text) == 2
    ), "irregular text (['hello ' * 600]) should result in 2 texts"



"""PromptTemplate abstract class along with the instances of the class."""

import re
from abc import abstractmethod
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

from confection import registry
from spacy.tokens import Doc

from conspiracies.docprocessing.relationextraction.data_classes import (
    SpanTriplet,
    StringTriplet,
)


class PromptTemplate:
    """Abstract class for template classes.

    The class is used to create a template for the triplet extraction
    task. The template is used to create a prompt for generative models
    such as GPT-3 and chatGPT.
    """

    def __init__(
        self,
        examples: List[Doc],
        task_description: Optional[str] = None,
    ):
        """
        Args:
            examples: A tuple of a spacy Docs
            task_description: A description of the task
        """

        self.examples = examples
        if task_description:
            self.task_description = task_description
        else:
            self.task_description = "Extract triplets on the form \
(Subject - predicate - object) from the following tweet. \
First, you will see a few examples."

    @abstractmethod
    def create_prompt(self, target: str) -> Union[str, List[Dict[str, str]]]:
        """Create a prompt based on the task description and examples and
        target."""
        pass

    @abstractmethod
    def parse_prompt(self, prompt: str, target_tweet: str) -> List[StringTriplet]:
        """Parse a prompt into a target tweet and triplets."""
        pass

    def set_examples(self, examples: List[Doc]):
        """Updates examples."""
        self.examples = examples


@registry.prompt_templates("conspiracies/template_1")
class PromptTemplate1(PromptTemplate):
    def __init__(
        self,
        examples: List[Doc],
        task_description: Optional[str] = None,
    ):
        super().__init__(task_description=task_description, examples=examples)
        if not task_description:
            self.task_description = """Extract semantic triplets from the following tweet. 
The semantic triplets should be on the form (Subject - Verb Phrase - Object), where the verb phrase includes all particles and modifyers. 
There should always be exactly three elements in a triplet, no more no less. 
They should be presented with each element in parentheses as shown below:"""  # noqa: E501

    def create_prompt(
        self,
        target: str = "",
    ) -> str:
        """Create a prompt template on the form

        ```
        {task_description}

        Tweet: {tweet1}
        Triplet: {triplets1}
        ...
        Tweet: {tweetN}
        Triplet: {tripletsN}

        Tweet: {target_tweet}
        ```

        Args:
            target: The tweet to be annotated.
        """
        examples_str = "---\n\n"
        for tweet in self.examples:
            triplets = tweet._.relation_triplets
            examples_str += f"Tweet: {tweet.text}\n"

            if len(triplets) == 0:
                examples_str += "Triplets: Null \n\n"
            else:
                for i, triplet in enumerate(triplets):
                    triplet_str = "({subj}) ({pred}) ({obj})"
                    triplet_str = triplet_str.format(
                        subj=triplet.subject.text,
                        pred=triplet.predicate.text,
                        obj=triplet.object.text,
                    )
                    if i == 0:
                        examples_str += f"Triplets: {triplet_str}\n"
                    else:
                        examples_str += f"\t{triplet_str}\n"

            examples_str += "---\n"
            tweet_string = f"{self.task_description}\n\n{examples_str}\nTweet: "
            prompt = tweet_string

        return prompt + f"{target}\nTriplets:"

    @staticmethod
    def create_string_triplets(tweet, triplet):
        subject, predicate, object = triplet
        return StringTriplet(
            subject=subject,
            predicate=predicate,
            object=object,
            text=tweet,
        )

    def parse_prompt(
        self,
        prompt_response: str,
        target_tweet: str,
    ) -> List[StringTriplet]:
        """Parse a prompt into a target tweet and triplets."""

        def strip_triplet(triplet):
            return [
                elem
                for elem in [element.replace(")", "").strip() for element in triplet]
                if elem not in ["", " "]
            ]

        data = [row for row in prompt_response.split("\n") if row not in ["", "---"]]
        triplets = []
        for line in data:
            tmp_triplet = line.split("(")
            triplet = strip_triplet(tmp_triplet)
            if len(triplet) == 3:
                triplets.append(self.create_string_triplets(target_tweet, triplet))
        return triplets


@registry.prompt_templates("conspiracies/template_2")
class PromptTemplate2(PromptTemplate):
    def __init__(
        self,
        examples: List[Doc],
        task_description: Optional[str] = None,
    ):
        super().__init__(task_description=task_description, examples=examples)
        if not task_description:
            self.task_description = """Extract semantic triplets from the following tweet. 
The semantic triplets should be on the form (Subject - Verb Phrase - Object), where the verb phrase includes all particles and modifyers. 
There should always be exactly three elements in a triplet, no more no less. 
They should be presented with each element in parentheses as shown below:"""  # noqa: E501

    def create_prompt(
        self,
        target: str = "",
    ) -> str:
        """Create a prompt template on the form '''.

        {task_description}
        Tweet: {tweet1}
        Tweet: {tweet2}
        ...
        Tweet: {tweetN}

        Triplet: {triplets1}
        Triplet: {triplets2}
        ...
        Triplet: {tripletsN}

        Tweet: {target_tweet}
        '''
        """
        tweets_block = "---\n\n"
        tweets = self.examples
        tweet_triplets = [tweet._.relation_triplets for tweet in tweets]
        for tweet in tweets:
            tweets_block += f"Tweet: {tweet.text}\n\n"

        triplet_block = ""
        for triplets in tweet_triplets:
            if len(triplets) == 0:
                triplet_block += "Triplets:\n"
                continue
            for i, triplet in enumerate(triplets):  # type: ignore
                triplet_str = "({subj}) ({pred}) ({obj})"
                triplet_str = triplet_str.format(
                    subj=triplet.subject.text,
                    pred=triplet.predicate.text,
                    obj=triplet.object.text,
                )
                if i == 0:
                    triplet_block += f"Triplets: {triplet_str}\n"
                else:
                    triplet_block += f"\t{triplet_str}\n"
        triplet_block += "\n---\n\n"
        prompt = f"{self.task_description}\n\n{tweets_block}\n{triplet_block}"

        return prompt + f"Tweet: {target}\n\nTriplets:"

    @staticmethod
    def create_string_triplets(tweet, triplet):
        subject, predicate, object = triplet
        return StringTriplet(
            subject=subject,
            predicate=predicate,
            object=object,
            text=tweet,
        )

    def parse_prompt(
        self,
        prompt_response: str,
        target_tweet: str,
    ) -> List[StringTriplet]:
        """Parse a prompt into a target tweet and triplets."""

        def strip_triplet(triplet):
            return [
                elem
                for elem in [element.replace(")", "").strip() for element in triplet]
                if elem not in ["", " "]
            ]

        data = [row for row in prompt_response.split("\n") if row not in ["", "---"]]
        triplets = []
        for line in data:
            tmp_triplet = line.split("(")
            triplet = strip_triplet(tmp_triplet)
            if len(triplet) == 3:
                triplets.append(self.create_string_triplets(target_tweet, triplet))
        return triplets


@registry.prompt_templates("conspiracies/markdown_template_1")
class MarkdownPromptTemplate1(PromptTemplate):
    def __init__(
        self,
        examples: List[Doc],
        task_description: Optional[str] = None,
    ):
        super().__init__(task_description=task_description, examples=examples)
        if not task_description:
            self.task_description = """Extract semantic triplets from the following tweet. 
The semantic triplets should be on the form (Subject - Verb Phrase - Object), where the verb phrase includes all particles and modifyers. 
There should always be exactly three elements in a triplet, no more no less. 
They should be put in a markdown table as shown below:"""  # noqa: E501

    def create_prompt(
        self,
        target: str = "",
    ) -> str:
        """Create a prompt template on the form

        ```
        {task_description}

        | Tweet | Subject | Predicate | Object |
        | --- | --- | --- | --- |
        | {tweet 1} | {subject 1} | {predicate 1} | {object 1} |
        |           | {subject 2} | {predicate 2} | {object 2} |
        ...
        | {tweet n} | {subject 1} | {predicate 1} | {object 1} |
        |           | {subject 2} | {predicate 2} | {object 2} |
        | {target_tweet} |
        ```
        """
        tweet_string = f"{self.task_description}\n| Tweet | Subject | Predicate | Object |\n| --- | --- | --- | --- |"  # noqa: E501
        for example in self.examples:
            triplets = example._.relation_triplets
            if len(triplets) == 0:
                tweet_string += f"\n| {example} | | | |"
                continue
            for i, triplet in enumerate(triplets):
                subj = triplet.subject.text
                pred = triplet.predicate.text
                obj = triplet.object.text

                if i == 0:
                    tweet_string += f"\n| {example} | {subj} | {pred} | {obj} |"
                else:
                    tweet_string += f"\n| | {subj} | {pred} | {obj} |"
        prompt = tweet_string

        return prompt + f"\n| {target} |"

    @staticmethod
    def create_string_triplets(tweet, triplet):
        subject, predicate, object = triplet
        return StringTriplet(
            subject=subject,
            predicate=predicate,
            object=object,
            text=tweet,
        )

    def parse_prompt(
        self,
        prompt_response: str,
        target_tweet: str,
    ) -> List[StringTriplet]:
        """Parse a prompt into a target tweet and triplets."""
        text = prompt_response.replace(self.task_description, "")
        data = [row for row in text.split("\n") if row != ""]
        triplets: List[StringTriplet] = []

        for row in data:
            elements = [
                element.strip()
                for element in row.split("|")
                if element not in ["", " "]
            ]
            # If the response has added another tweet, return triplets as is now
            if len(elements) == 4:
                return triplets
            # If there are only three elements they are triplets for the current tweet
            elif len(elements) == 3:
                as_string_triplet = self.create_string_triplets(target_tweet, elements)
                triplets.append(as_string_triplet)

        return triplets


@registry.prompt_templates("conspiracies/markdown_template_2")
class MarkdownPromptTemplate2(PromptTemplate):
    def __init__(
        self,
        examples: List[Doc],
        task_description: Optional[str] = None,
    ):
        super().__init__(task_description=task_description, examples=examples)
        if not task_description:
            self.task_description = """Extract semantic triplets from the following tweet. 
The semantic triplets should be on the form (Subject - Verb Phrase - Object), where the verb phrase includes all particles and modifyers. 
There should always be exactly three elements in a triplet, no more no less. 
They should be put in a markdown table as shown below:"""  # noqa: E501

    def create_prompt(
        self,
        target: str = "",
    ) -> str:
        """Create a prompt template on the form.

        ```
        {Task description}

        {tweet 1}

        | Subject | Predicate | Object |
        | ---- | ---- | ---- |
        | {triplet 1 from tweet 1}
        | {triplet 2 from tweet 1}

        {tweet 2}

        | Subject | Predicate | Object |
        | ---- | ---- | ---- |
        | {triplet 1 from tweet 2}
        | {triplet 2 from tweet 2}
        ...

        {target tweet}

        | Subject | Predicate | Object |
        | ---- | ---- | ---- |
        ```
        """
        header = "| Subject | Predicate | Object |\n| --- | --- | --- |"

        tweet_string = f"{self.task_description}\n\n"
        for example in self.examples:
            triplets = example._.relation_triplets
            tweet_string += example.text + "\n\n" + header
            if len(triplets) == 0:
                tweet_string += "\n\n"
                continue
            for triplet in triplets:
                subj = triplet.subject.text
                pred = triplet.predicate.text
                obj = triplet.object.text

                tweet_string += f"\n| {subj} | {pred} | {obj} |"
            tweet_string += "\n\n"
        prompt = tweet_string

        return prompt + target + "\n\n" + header + "\n"

    @staticmethod
    def create_string_triplets(tweet, triplet):
        subject, predicate, object = triplet
        return StringTriplet(
            subject=subject,
            predicate=predicate,
            object=object,
            text=tweet,
        )

    def parse_prompt(
        self,
        prompt_response: str,
        target_tweet: str,
    ) -> List[StringTriplet]:
        """Parse a prompt into a target tweet and triplets.

        Ignores extracted triplets that do not contain exactly three
        elements
        """
        # Remove task description
        text = prompt_response.replace(self.task_description, "")
        data = [row for row in text.split("\n") if row != ""]
        triplets: List[StringTriplet] = []

        for row in data:
            elements = [element.strip() for element in row.split("|") if element != ""]
            # If the response has added another tweet, return triplets as is now
            if len(elements) == 4:
                return triplets
            # If there are only three elements they are triplets for the current tweet
            elif len(elements) == 3:
                as_string_triplet = self.create_string_triplets(target_tweet, elements)
                triplets.append(as_string_triplet)

        return triplets


@registry.prompt_templates("conspiracies/xml_style_template")
class XMLStylePromptTemplate(PromptTemplate):
    def __init__(
        self,
        examples: List[Doc],
        task_description: Optional[str] = None,
        tags: List[str] = ["subject", "predicate", "object"],
    ):
        super().__init__(task_description=task_description, examples=examples)
        self.tags = tags
        if not task_description:
            self.task_description = """Tag the following tweet with triplets using HTML tags.
Semantic triplets consists of the elements subject, verb phrase and object. The verb phrase includes all particles and modifyers.
There should always be exactly three elements in a triplet, no more no less.
The subject is enclosed between <subject-n> and </subject-n>, the verb phrase between <predicate-n> and </predicate-n> and the object between <object-n> and </object-n>.
n is the number of the triplet, starting at 1. Elements of one triplet can be contained within elements of another triplet.
The triplets should be tagged in the tweet as shown below:"""  # noqa: E501

    @staticmethod
    def create_xml_example(doc: Doc, triplets: List[SpanTriplet]) -> str:
        """This function creates an XML-like prompt from a spacy doc and a list
        of triplets. It inserts XML tags around the subject, predicate and
        object of each triplet based on the start and end index of the span.

        Args:
            doc: The spacy doc to insert the html tags into.
            triplets: The triplets specifying the subjects, predicates and
                objects.

        Returns:
            The doc with html tags around the subjects, predicates and objects of
                each triplet
        """
        result_string = ""
        for i, d in enumerate(doc):
            for n, triplet in enumerate(triplets, 1):
                if triplet.subject.end == i:
                    result_string += f"</subject-{n}>"
                if triplet.object.end == i:
                    result_string += f"</object-{n}>"
                if triplet.predicate.end == i:
                    result_string += f"</predicate-{n}>"

                if triplet.subject.start == i:
                    result_string += f"<subject-{n}>"
                if triplet.object.start == i:
                    result_string += f"<object-{n}>"
                if triplet.predicate.start == i:
                    result_string += f"<predicate-{n}>"
            result_string += d.text_with_ws
        return result_string

    def create_prompt(
        self,
        target: str = "",
    ) -> str:
        """Create a prompt template on the form.

        ```
        {task_description}

        {tweet 1}
        {xml_tagged tweet 1}

        ...

        {tweet n}
        {xml_tagged tweet n}

        {target tweet}
        """
        prompt = f"{self.task_description}\n\n"
        for doc in self.examples:
            triplets = doc._.relation_triplets
            prompt += doc.text + "\n" + self.create_xml_example(doc, triplets) + "\n\n"

        return prompt + target + "\n"

    @staticmethod
    def __parse(text: str, tags=["subject", "predicate", "object"]) -> Dict[str, Any]:
        extract_tags = XMLStylePromptTemplate.extract_tags
        remove_tags = XMLStylePromptTemplate.remove_tags
        update_span = XMLStylePromptTemplate.update_span

        tags_info = extract_tags(text, tags)

        open_tags = {}
        closed_tags = {}
        for tag_info in tags_info:
            n = tag_info["n"]
            tag = tag_info["tag"]
            if tag_info["is_close"]:
                closed_tags[f"{tag}-{n}"] = {"span": tag_info["span"]}
            else:
                open_tags[f"{tag}-{n}"] = {"span": tag_info["span"], "tag": tag, "n": n}

        # extract triplets
        triplets = defaultdict(  # type: ignore
            lambda: {"subject": None, "predicate": None, "object": None},
        )
        spans_to_update = []
        for tag_n in open_tags:
            if tag_n in closed_tags:
                content_span = list(
                    (open_tags[tag_n]["span"][1], closed_tags[tag_n]["span"][0]),
                )
                content = text[content_span[0] : content_span[1]]
                tag = open_tags[tag_n]["tag"]
                n = open_tags[tag_n]["n"]
                triplets[n][tag] = {  # type: ignore
                    "text": remove_tags(content),
                    "span": content_span,
                }
                spans_to_update.append(content_span)

        spans_to_remove = [tag_info["span"] for tag_info in tags_info]
        for span in spans_to_update:
            static_span = tuple(span)
            content = text[span[0] : span[1]]

            for span_to_remove in spans_to_remove:
                _span = update_span(static_span, span_to_remove)  # type: ignore
                span[0] += _span[0]
                span[1] += _span[1]

        return {"text": remove_tags(text), "triplets": triplets}

    def parse_prompt(
        self,
        prompt_response: str,
        target_tweet: str,
    ) -> List[StringTriplet]:
        """Parse a prompt into a target tweet and triplets.

        Does not use the target_tweet, but required for consistency
        """
        parse = self.__parse(prompt_response, tags=self.tags)

        def to_string_triplet(triplet):
            return StringTriplet(
                subject=triplet["subject"]["text"],
                predicate=triplet["predicate"]["text"],
                object=triplet["object"]["text"],
                subject_char_span=triplet["subject"]["span"],
                predicate_char_span=triplet["predicate"]["span"],
                object_char_span=triplet["object"]["span"],
                text=parse["text"],
            )

        triplets = []

        for n, triplet in parse["triplets"].items():
            # NB Shouldnt this be taken care of earlier??
            if (
                triplet["subject"] is not None
                and triplet["predicate"] is not None
                and triplet["object"] is not None
            ):
                str_triplet = to_string_triplet(triplet)
                triplets.append(str_triplet)
        return triplets

    @staticmethod
    def extract_tags(
        text: str,
        tags: List[str] = ["subject", "predicate", "object"],
    ) -> List[dict]:
        """Extractss all tags on the form: <{tag}-{n}>{content}</{tag}-{n}>

        and returns a dict on the form:
        {tag: tag, span: (start, end), n: n, is_close: bool}
        """
        # create regex pattern to match both start and close tags
        pattern = r"<[/]?(" + "|".join(tags) + r")-\d+>"

        def extract_tag_info(match):
            tag = match.group()[1:-1]
            is_close = tag.startswith("/")
            if is_close:
                tag = tag[1:]
            tag, n = tag.split("-")
            return {"tag": tag, "span": match.span(), "n": n, "is_close": is_close}

        return [extract_tag_info(m) for m in re.finditer(pattern, text)]

    @staticmethod
    def remove_tags(text: str) -> str:
        """Remove all tags from a string."""
        return re.sub(r"<[/]?\w+-\d+>", "", text)

    @staticmethod
    def update_span(
        span: Tuple[int, int],
        deleted_span: Tuple[int, int],
    ) -> Tuple[int, int]:
        """Updates a span by after a part of the text has been deleted."""
        diff = deleted_span[1] - deleted_span[0]

        # if deleted span is before span
        if span[0] >= deleted_span[1]:
            return -diff, -diff
        # if deleted span is after span
        if span[1] <= deleted_span[0]:
            return 0, 0
        # if deleted span is inside span
        if span[0] <= deleted_span[0] and span[1] >= deleted_span[1]:
            return 0, -diff
        # if deleted span is overlapping start of span
        if span[0] < deleted_span[0] and span[1] < deleted_span[1]:
            return 0, deleted_span[0] - span[1]
        # if deleted span is overlapping end of span
        if span[0] > deleted_span[0] and span[1] > deleted_span[1]:
            return deleted_span[1] - span[0], 0
        # if deleted span is equal to span
        if span[0] == deleted_span[0] and span[1] == deleted_span[1]:
            raise ValueError("Span to delete is equal to span to update")
        raise ValueError("Unknown error")


@registry.prompt_templates("conspiracies/chatgpt_style_template")
class chatGPTPromptTemplate(PromptTemplate):
    def __init__(
        self,
        examples: List[Doc],
        task_description: Optional[str] = None,
    ):
        super().__init__(task_description=task_description, examples=examples)
        if not task_description:
            self.task_description = """I need you to extract semantic triplets from tweets. 
The semantic triplets should be on the form (Subject - Verb Phrase - Object), where the verb phrase includes all particles and modifyers. 
There should always be exactly three elements in a triplet, no more no less, and they should be presented in a markdown table.
First, I will provide you with a few examples."""  # noqa: E501

    def create_prompt(self, target: str) -> List[Dict[str, str]]:
        """Create a prompt for the chatGPT model. Form is:
        [
        {"role": "system",
         "content": "You are a helpful assistant who is also an expert linguist. "
        },
        {"role": "user",
         "content": "Hi, I need your help on a linguistic question."
        },
        {"role": "assistant",
         "content": "Sure, what is the task?"
        },
        {"role": "user",
         "content": task_description
        },
        {"role": "assistant",
         "content": "That sounds like something I can help you with. Let me see the examples."
        },
        {"role": "user",
            "content": "The tweet is:\n{example_1}\nThe triplets in the tweet are:{triplet_1\ntriplets_2}\n"
        },
        {"role": "assistant",
        "content": "Yes I see! What is the next tweet and triplets?\n"
        },

        ...

        {"role": "user",
         "content": "The tweet is:\n{example_n}\nThe triplets in the tweet are:{triplet_1\n ... triplet_m}\n"
        },
        {"role": "user",
         "content": 'All right, now you try. The tweet is:\n{target}\nWhat are the triplets?'
        },
        ]
        """  # noqa: E501
        message_dicts = [
            {
                "role": "system",
                "content": "You are a helpful assistant \
who is also an expert linguist. ",
            },
            {
                "role": "user",
                "content": "Hi, I need your help on a linguistic question.",
            },
            {
                "role": "assistant",
                "content": "Sure, what is the task?",
            },
            {
                "role": "user",
                "content": self.task_description,
            },
            {
                "role": "assistant",
                "content": "That sounds like something I can help you with. \
Let me try the first example!",
            },
        ]
        for i, example in enumerate(self.examples):
            triplets = example._.relation_triplets
            if i == 0:
                message_dicts.append(
                    {
                        "role": "user",
                        "content": f"The tweet is:\n{example.text}\n",
                    },
                )
            else:
                message_dicts.append(
                    {
                        "role": "user",
                        "content": f"Correct! The next tweet is:\n{example.text}\n",
                    },
                )
            message_string = "I extracted the following triplets:\n"
            for triplet in triplets:
                message_string += (
                    f"\t{triplet.subject} - {triplet.predicate} - {triplet.object}\n"
                )
            message_dicts.append(
                {
                    "role": "assistant",
                    "content": message_string,
                },
            )
        message_dicts.append(
            {
                "role": "user",
                "content": f"The tweet is:\n{target}\n",
            },
        )
        return message_dicts

    @staticmethod
    def create_string_triplets(tweet, triplet):
        subject, predicate, object = triplet
        return StringTriplet(
            subject=subject,
            predicate=predicate,
            object=object,
            text=tweet,
        )

    def parse_prompt(
        self,
        prompt_response: str,
        target_tweet: str,
    ) -> List[StringTriplet]:
        """Parse a prompt into a target tweet and triplets.

        Ignores extracted triplets that do not contain exactly three
        elements
        """

        triplets: List[StringTriplet] = []
        lines = [
            line.strip()
            for line in prompt_response.split("\n")
            if line not in ["", " "]
        ]
        for line in lines:
            if "-" in line:
                extractions = [element.strip() for element in line.split("-")]
                if len(extractions) == 3:
                    triplets.append(
                        self.create_string_triplets(target_tweet, extractions),
                    )
        return triplets



from pathlib import Path
from typing import List, Union, Iterable, Tuple

import jsonlines
from spacy.language import Language
from spacy.tokens import Doc

from conspiracies.docprocessing.relationextraction.data_classes import (
    install_extensions,
)
from conspiracies.docprocessing.relationextraction.gptprompting import (
    DocTriplets,
    SpanTriplet,
)


def _doc_to_json(doc: Union[Doc, Tuple[Doc, str]], include_span_heads=True):
    if isinstance(doc, Tuple):
        doc, id_ = doc
    else:
        id_ = None
    if Doc.has_extension("relation_triplets"):
        triplets = doc._.relation_triplets
    else:
        triplets = []
    json = doc.to_json()
    if id_ is not None:
        json["id"] = id_
    json["semantic_triplets"] = [
        triplet.to_dict(include_doc=False, include_span_heads=include_span_heads)
        for triplet in triplets
    ]
    return json


def _doc_from_json(json: dict, nlp: Language) -> Doc:
    doc = Doc(nlp.vocab).from_json(json)
    triplets = [
        SpanTriplet.from_dict(triplet_json, nlp=nlp, doc=doc)
        for triplet_json in json["semantic_triplets"]
    ]
    if not Doc.has_extension("relation_triplets"):
        install_extensions()
    doc._.relation_triplets = DocTriplets(span_triplets=triplets, doc=doc)
    return doc


def docs_to_jsonl(
    docs: Iterable[Union[Doc, Tuple[Doc, str]]],
    path: Union[Path, str],
    append=False,
    include_span_heads=True,
) -> None:
    """Write docs and triplets to a jsonl file.

    Args:
        docs: a list of docs. If the docs have the extension
            "relation_triplets", the triplets will written to the jsonl file.
        path: path to the jsonl file.
        append: whether to append to file instead of overwriting
        include_span_heads: whether to output an "extracted_head" field in the JSON
            output from :class:`HeadWordExtractionComponent`
    """
    with jsonlines.open(path, "a" if append else "w") as writer:
        writer.write_all(
            _doc_to_json(doc, include_span_heads=include_span_heads) for doc in docs
        )


def docs_from_jsonl(
    path: Union[Path, str],
    nlp: Language,
) -> List[Doc]:
    """Read docs and triplets from a jsonl file.

    Args:
        path: path to the jsonl file.
        nlp: a spacy language model.

    Returns:
        A list of docs with the extension `doc._.relation_triplets` set.
    """
    if not Doc.has_extension("relation_triplets"):
        install_extensions(force=True)
    docs = []
    with jsonlines.open(path, "r") as reader:
        for json in reader:
            doc = _doc_from_json(json, nlp)
            docs.append(doc)
    return docs



from pathlib import Path

import pytest

from conspiracies.pipeline.config import (
    PipelineConfig,
    BaseConfig,
    PreProcessingConfig,
    DocProcessingConfig,
    CorpusProcessingConfig,
)


@pytest.fixture()
def path():
    data_path = Path(__file__).parent / "test_data" / "test_config.toml"
    return data_path


def test_config_loading(path: str):
    config = PipelineConfig.from_toml_file(path)
    assert config == PipelineConfig(
        base=BaseConfig(
            project_name="test",
            output_root="output",
            language="en",
        ),
        preprocessing=PreProcessingConfig(
            enabled=False,
            input_path="some_input",
            doc_type="test",
            metadata_fields={"metadata_field"},
            extra={
                "some_extra_value": 1234,
            },
        ),
        docprocessing=DocProcessingConfig(
            triplet_extraction_method="test",
        ),
        corpusprocessing=CorpusProcessingConfig(),
    )



"""Utilities for downloading models."""

import os
import shutil
from pathlib import Path
from typing import Union

from tqdm import tqdm
from wasabi import msg

MODEL_URLS = {
    "da_coref_twitter_v1": {
        "url": "https://sciencedata.dk/shared/"
        + "f20703362e0a771f004f8ff2026afb0f?download",
        "folder_name": "da_coref_twitter_v1",
    },
}
DEFAULT_CACHE_DIR = os.getenv(
    "CONSPIRACIES_CACHE_DIR",
    os.path.join(str(Path.home()), ".conspiracies"),
)


class DownloadProgressBar(tqdm):
    def update_to(self, b: int = 1, bsize: int = 1, tsize=None) -> None:
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url: str, output_path: str) -> None:
    import urllib.request

    with DownloadProgressBar(
        unit="B",
        unit_scale=True,
        miniters=1,
        desc=url.split("/")[-1],
    ) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


def download_model(
    model: str,
    save_path: Union[Path, str, None] = None,
    force: bool = False,
    verbose: bool = True,
    open_unverified_connection: bool = False,
) -> str:
    """Download a model name from list of models.

    Args:
        model(str): Name of the model to download.
        save_path(Union[Path, str, None], optional): Path to save the model to.
            If None, the model will be saved to the default cache directory.
        force(bool, optional): If True, the model will be downloaded even if it
            already exists. Defaults to False.
        verbose(bool, optional): If True, the model will be downloaded and saved
            to the default cache directory. Defaults to True.
        open_unverified_connection (bool, optional): Should you download from an
            unverified connection. Defaults to False.

    Returns:
        str: Path to the downloaded model.
    """

    if open_unverified_connection:
        import ssl

        ssl._create_default_https_context = ssl._create_unverified_context

    if save_path is None:
        save_path = DEFAULT_CACHE_DIR

    if model not in MODEL_URLS:
        raise ValueError(
            "The model is not available."
            + f" Models available include: {list(MODEL_URLS.keys())}",
        )
    url = MODEL_URLS[model]["url"]
    path = os.path.join(save_path, MODEL_URLS[model]["folder_name"])

    # only download model if required
    if os.path.exists(path) and force is False:
        return path

    dl_path = os.path.join(save_path, "tmp.zip")
    Path(save_path).mkdir(parents=True, exist_ok=True)

    if verbose is True:
        msg.info(f"\nDownloading '{model}'")
    download_url(url, dl_path)

    shutil.unpack_archive(dl_path, save_path)  # unzip
    os.remove(dl_path)

    if verbose is True:
        msg.info(f"Model successfully downloaded to  {path}")
    return path



import numpy as np
from conspiracies.docprocessing.relationextraction.gptprompting import (
    DocTriplets,
    score_open_relations,
)
from spacy.training import Example

from .utils import docs_with_triplets  # noqa F401


def test_prompt_relation_evaluate(docs_with_triplets):  # noqa: F811
    # TODO: the document references in Span and Triple objects go all over the place
    #  through this test

    docs = docs_with_triplets

    keys = [
        "exact_span_match",
        "exact_string_match",
        "normalized_span_overlap",
        "normalized_string_overlap",
    ]

    # # perfect match
    doc = docs[0]
    example = Example(doc, doc)
    score = score_open_relations([example])

    for key in keys:
        assert score[f"{key}_f1"] == 1.0
        assert score[f"{key}_precision"] == 1.0
        assert score[f"{key}_recall"] == 1.0

    # no match
    doc_no_triplets = doc[:].as_doc()
    doc_no_triplets._.relation_triplets = DocTriplets(
        span_triplets=[],
        doc=doc_no_triplets,
    )
    example = Example(reference=doc, predicted=doc_no_triplets)
    score = score_open_relations([example])

    assert score["n_predictions"] == 0
    assert score["n_references"] == 3
    for key in keys:
        assert np.isnan(score[f"{key}_f1"])
        assert np.isnan(score[f"{key}_precision"])
        assert score[f"{key}_recall"] == 0.0

    # partial match
    doc_partial_match = doc[:].as_doc()
    triplet = doc._.relation_triplets[0]
    doc_partial_match._.relation_triplets = DocTriplets(
        span_triplets=[triplet],
        doc=doc_partial_match,
    )

    example = Example(reference=doc, predicted=doc_partial_match)
    score = score_open_relations([example])

    for key in keys:
        assert 1 > score[f"{key}_f1"] > 0
        assert score[f"{key}_recall"] - 0.333 < 0.01
        assert score[f"{key}_precision"] == 1

    # partial match (overmatched)
    example = Example(predicted=doc, reference=doc_partial_match)
    score = score_open_relations([example])

    for key in keys:
        assert 1 > score[f"{key}_f1"] > 0
        assert score[f"{key}_recall"] == 1
        assert score[f"{key}_precision"] - 0.333 < 0.01



import logging
from typing import Dict, Iterable, Iterator, List

import spacy
from spacy import Vocab
from spacy.language import Language
from spacy.pipeline import TrainablePipe
from spacy.tokens import Doc
from spacy_transformers.align import get_alignment
from transformers import AutoTokenizer

from .knowledge_triplets import KnowledgeTriplets
from .multi2oie_utils import (
    match_extraction_spans_to_wp,
    wp2tokid,
    wp_span_to_token,
)
from conspiracies.docprocessing.relationextraction.data_classes import (
    install_extensions,
    DocTriplets,
)


class SpacyRelationExtractor(TrainablePipe):
    """SpaCy pipeline component that adds a multilingual relation-extraction
    component. The extractions are saved in the doc._.relation_triplets,

    ._.relation_head,

    ._.relation_relation, and ._.relation_tail attributes.
    Args:
        vocab (Vocab): The Vocab object for the pipeline.
        confidence_threshold (float): A threshold for model confidence to filter
            uncertain relations by.
        model_args (Dict): Keyword arguments for KnowledgeTriplets
            (e.g. batch_size, path)
        name (str): spaCy internal
        labels (List(str)): Required for TrainablePipe but unused. Leave as empty
            list (or don't)
    """

    def __init__(
        self,
        vocab: Vocab,
        name: str,
        labels: List[str],
        confidence_threshold: float,
        model_args,
    ):
        self.name = name
        self.vocab = vocab
        self.model = KnowledgeTriplets(**model_args)
        self.tokenizer = AutoTokenizer.from_pretrained("bert-base-multilingual-cased")
        self.confidence_threshold = confidence_threshold

        install_extensions()

        self.ignore_tokenizers_warning()
        self.ignore_transformers_warnings()

        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def ignore_tokenizers_warning() -> None:
        """Ignore the warning from the transformers library caused by forking
        the process."""
        import os

        os.environ["TOKENIZERS_PARALLELISM"] = "false"

    def ignore_transformers_warnings(self) -> None:
        from transformers import logging

        logging.set_verbosity_error()

    def set_annotations(self, doc: Iterable[Doc], predictions: Dict) -> None:
        """Assign the extracted features to the Doc objects. Extractions below
        the confidence threshold are filtered, wordpieces and spacy tokens are
        aligned and then attributes are set.

        Args:
            docs (Iterable[Doc]): The documents to modify.
            predictions: (Dict): A batch of outputs from
                KnowledgeTriplets.extract_relations().
        """
        # Many things can go wrong in here every now and then, and a pipeline of many
        # documents can break down. Until those things get fixed, the whole method is
        # wrapped in a broad try/except statement
        try:
            self.do_set_annotations(doc, predictions)
        except Exception as e:
            self.logger.exception(e)

    def do_set_annotations(self, doc: Iterable[Doc], predictions: Dict) -> None:
        # get nested list of indices above confidence threshold
        filtered_indices = [
            [
                idx
                for idx, conf in enumerate(confidences)
                if conf > self.confidence_threshold
            ]
            for confidences in (predictions["confidence"])
        ]
        # only keep relations above the threshold
        for key in ["extraction", "extraction_span", "confidence"]:
            predictions[key] = [
                [values[filter_idx] for filter_idx in indices]
                for indices, values in zip(filtered_indices, predictions[key])
            ]

        # return if no extractions are above threshold
        if not predictions["extraction"]:
            return

        # concatenate wordpieces and concatenate extraction span. Handle new extraction
        #           # spans by adding the length of the previous
        # Join wordpieces to single list
        wordpieces = [j for i in predictions["wordpieces"] for j in i]
        # Concatenate extraction spans.

        matched_extractions = match_extraction_spans_to_wp(
            predictions["extraction_span"],
            predictions["wordpieces"],
        )

        align = get_alignment([doc], [wordpieces], self.tokenizer.all_special_tokens)
        wp2tokids = wp2tokid(align)
        aligned_extractions = wp_span_to_token(
            matched_extractions,
            wp2tokids,
            doc,  # type: ignore
        )

        # Set doc level attributes
        merged_confidence = [j for i in predictions["confidence"] for j in i]
        setattr(doc._, "relation_confidence", merged_confidence)  # type: ignore
        triplets = DocTriplets(span_triplets=aligned_extractions, doc=doc)
        setattr(doc._, "relation_triplets", triplets)  # type: ignore

    def pipe(self, stream: Iterable[Doc], *, batch_size: int = 128) -> Iterator[Doc]:
        """Apply the pipe to a stream of documents.

        This usually happens under
        the hood when the nlp object is called on a text and all components are
        applied to the Doc. Batch size is controlled by `batch_size` when
        instatiating the nlp.pipe object.
        stream (Iterable[Doc]): A stream of documents.
        batch_size (int): The number of documents to buffer.
        YIELDS (Doc): Processed documents in order.
        DOCS: https://spacy.io/api/transformer#pipe
        """
        for doc in stream:
            sents = [sent.text for sent in doc.sents]
            self.set_annotations(doc, self.predict(sents))  # type: ignore
            yield doc

    def predict(self, docs: Iterable[Doc]) -> Dict:
        """Apply the pipeline's model to a batch of docs, without modifying
        them.

        Returns the extracted features as the FullTransformerBatch dataclass.
        docs (Iterable[Doc]): The documents to predict.
        RETURNS (Dict): The extracted features.
        DOCS: https://spacy.io/api/transformer#predict
        """
        return self.model.extract_relations(docs)  # type: ignore

    def __call__(self, doc: Doc) -> Doc:
        """Apply the pipe to one document.

        The document is modified in place,
        and returned. This usually happens under the hood when the nlp object
        is called on a text and all components are applied to the Doc.
        docs (Doc): The Doc to process.
        RETURNS (Doc): The processed Doc.
        DOCS: https://spacy.io/api/transformer#call
        """
        sents = [sent.text for sent in doc.sents]
        outputs = self.predict(sents)  # type: ignore
        self.set_annotations(doc, outputs)  # type: ignore
        return doc


@Language.factory(
    "relation_extractor",
    default_config={
        "confidence_threshold": 2.5,
        "labels": [],
        "model_args": {
            "batch_size": 64,
        },
    },
)
def make_relation_extractor(
    nlp: Language,
    name: str,
    confidence_threshold: float,
    labels: List,
    model_args: Dict,
):
    return SpacyRelationExtractor(
        nlp.vocab,
        name=name,
        confidence_threshold=confidence_threshold,
        labels=labels,
        model_args=model_args,
    )


if __name__ == "__main__":
    nlp = spacy.load("da_core_news_sm")

    test_sents = [
        "Pernille Blume vinder delt EM-sølv i Ungarn.",
        "Pernille Blume blev nummer to ved EM på langbane i disciplinen 50 meter fri.",
        "Hurtigst var til gengæld hollænderen Ranomi Kromowidjojo, der sikrede sig "
        + "guldet i tiden 23,97 sekunder.",
        "Og at formen er til en EM-sølvmedalje tegner godt, siger Pernille Blume med"
        + " tanke på, at hun få uger siden var smittet med corona.",
        "Ved EM tirsdag blev det ikke til medalje for den danske medley for mixede "
        + "hold i 4 x 200 meter fri.",
        "In a phone call on Monday, Mr. Biden warned Mr. Netanyahu that he could "
        + "fend off criticism of the Gaza strikes for only so long, according to two"
        + " people familiar with the call",
        "That phone call and others since the fighting started last week reflect Mr."
        + " Biden and Mr. Netanyahu’s complicated 40-year relationship.",
        "Politiet skal etterforske Siv Jensen etter mulig smittevernsbrudd.",
        "En av Belgiens mest framträdande virusexperter har flyttats med sin familj "
        + "till skyddat boende efter hot från en beväpnad högerextremist.",
    ]

    config = {"confidence_threshold": 1.0, "model_args": {"batch_size": 10}}

    # model = KnowledgeTriplets()

    nlp.add_pipe("relation_extractor", config=config)

    pipe = nlp.pipe(test_sents)

    for d in pipe:
        print(d.text, "\n", d._.relation_triplets)



"""Extract examples of tweets and triplets. Examples are on ucloud in
data/triplet-extraction-gpt/tagged/ and on Grundtvig in.

/data/conspiracies/triplet-extraction-gpt
"""

import random
import re
from typing import List, Tuple
from spacy.tokens import Doc
from conspiracies.docprocessing.relationextraction.data_classes import (
    DocTriplets,
)


def has_multiple_triplets(spacy_triplets: DocTriplets):
    return len(spacy_triplets) > 1


def has_no_triplets(spacy_triplets: DocTriplets):
    return len(spacy_triplets) == 0


def has_one_mention(text_doc: Doc):
    return bool(re.search(r"@[\w]+", text_doc.text))


def has_multi_word_verb(spacy_triplets: DocTriplets):
    for triplet in spacy_triplets:
        if len(triplet.predicate) > 1:
            return True
    else:
        return False


def has_single_word_verb(spacy_triplets: DocTriplets):
    for triplet in spacy_triplets:
        if len(triplet.predicate) == 1:
            return True
    else:
        return False


def has_multi_word_subj(spacy_triplets: DocTriplets):
    for triplet in spacy_triplets:
        if len(triplet.subject) > 1:
            return True
    else:
        return False


def has_multi_word_obj(spacy_triplets: DocTriplets):
    for triplet in spacy_triplets:
        if len(triplet.object) > 1:
            return True
    else:
        return False


def criteria_sampling(
    criteria_keys: dict,
    n_target: int,
    input_examples: List[Doc],
):
    """This function tries to extract examples that fulfill the criteria, so
    that as many of the criteria as possible are fulfilled in the target
    tweets.

    If that is not possible, it will just sample randomly.
    """
    target_examples: List[Doc] = []
    while len(target_examples) < n_target:
        try:
            criteria = random.choice(list(criteria_keys.keys()))
            subset_function = criteria_keys[criteria]
            # function_key = "doc" if criteria == "has_one_mention" else "triplets"

            # Filter only examples that fulfill the criteria
            if criteria == "has_one_mention":
                useful_examples = list(
                    filter(lambda x: subset_function(x), input_examples),
                )
            else:
                useful_examples = list(
                    filter(
                        lambda x: subset_function(x._.relation_triplets),
                        input_examples,
                    ),
                )

            target = random.choice(useful_examples)
            criteria_keys.pop(criteria)

        except IndexError:
            target = random.choice(input_examples)

        target_examples.append(target)
        input_examples.remove(target)
    return target_examples, input_examples


def extract_examples(
    examples: List[Doc],
    n_target: int,
    cv: int = 1,
) -> Tuple[List[List[Doc]], List[List[Doc]]]:
    """This function extracts examples and targets for the triplet extraction
    task. It tries to extract examples that fulfill the criteria, so that as
    many of the criteria as possible are fulfilled in the target tweets. It can
    also be used to create cross validation sets.

    Args:
        examples (List[Doc]): tagged examples to extract from
        n_target (int): number of target examples to extract
        cv (int, optional): number of cv-sets to make. Defaults to 1,
            which means only one set of targets and examples are extracted.

    Returns:
        Tuple[List[List[Doc]], List[List[Doc]]]: _description_
    """
    # Manually create copy since spacy spans cannot be deepcopied
    example_dicts = [d for d in examples]

    target_list: List[Doc] = []
    example_list: List[Doc] = []
    for _ in range(cv):
        criteria_keys = {
            "has_multiple_triplets": has_multiple_triplets,
            "has_no_triplets": has_no_triplets,
            "has_one_mention": has_one_mention,
            "has_multi_word_verb": has_multi_word_verb,
            "has_single_word_verb": has_single_word_verb,
            "has_multi_word_subj1": has_multi_word_subj,
            "has_multi_word_obj": has_multi_word_obj,
        }
        extracted_targets, extracted_examples = criteria_sampling(
            criteria_keys,
            n_target,
            example_dicts,
        )
        example_list.append(
            [ex for prev_target in target_list for ex in prev_target]
            + extracted_examples,
        )
        target_list.append(extracted_targets)

    return target_list, example_list
